import logging
import threading

from ahc.Ahc import *
from ahc.Routing.STAR.STARNodeComponent import STARMessageTypes
from ahc.Routing.STAR.helper import MessageGenerator, STARTestBenchConfig

logger = logging.getLogger(__name__)


class ApplicationComponent(ComponentModel):
    def __init__(self, componentname, componentid):
        super(ApplicationComponent, self).__init__(componentname, componentid)
        self.generator = MessageGenerator(mps=STARTestBenchConfig.MPS, sender_fn=self.send_app_message)

    def on_init(self, eventobj: Event):
        threading.Timer(STARTestBenchConfig.WARM_UP, self.start_flow).start()

    def send(self, to, msg):
        message_header = GenericMessageHeader(STARMessageTypes.APP, self.componentinstancenumber, to)
        message = GenericMessage(message_header, GenericMessagePayload(msg))
        kickstarter = Event(self, EventTypes.MFRT, message)
        logger.debug(f"AppLayer {self.componentinstancenumber} sends message to AppLayer {to}")

        self.send_down(kickstarter)

    def on_message_from_bottom(self, eventobj: Event):
        message_from = eventobj.eventcontent.header.messagefrom
        payload = eventobj.eventcontent.payload.messagepayload

        logger.debug(f"AppLayer {self.componentinstancenumber} got message: {payload} from {message_from}")

    def start_flow(self):
        self.generator.start()

    def terminate(self):
        super().terminate()
        self.generator.terminate()

    def send_app_message(self):
        dest = self.componentinstancenumber
        neighbors = Topology().get_neighbors(self.componentinstancenumber)

        while dest == self.componentinstancenumber or dest in neighbors:
            dest = Topology().get_random_node().componentinstancenumber

        shortest_hop_count = len(Topology().allpairs_shortest_path()[self.componentinstancenumber][dest]) - 2
        hop_count = 0

        self.send(dest, {
            'text': 'Hello, World!',
            'from': self.componentinstancenumber,
            'to': dest,
            'shortest': shortest_hop_count,
            'hop_count': hop_count
        })



import threading
import time
from collections import Counter
from enum import Enum


class STARTestBenchConfig:
    TERMINATED = False
    SIMULATION_TIME = 10  # sec
    NODE_COUNT = 10
    DENSITY = 0.1  # probability of edge creation
    WARM_UP = 5  # sec
    MPS = 2  # message per second


class STARStatEvent(Enum):
    MSG_SENT = "MessageSent"
    UPDATE_MSG_SENT = "UpdateMessageSent"
    LSU_MSG_SENT = "LSUMessageSent"
    LSU_MSG_RECV = "LSUMessageSent"
    APP_MSG_SENT = "AppMessageSent"
    APP_MSG_RECV = "AppMessageReceived"
    LINK_UPDATED = "LinkUpdated"

    def __str__(self):
        return self.value

    def __repr__(self):
        return str(self)


class STARStats(object):
    def __init__(self):
        self.msg_sent = 0
        self.update_msg_sent = 0
        self.lsu_sent = 0
        self.lsu_recv = 0
        self.app_sent = 0
        self.app_recv = 0
        self.link_updated = 0
        self.total_shortest_hop_count = 0
        self.total_hop_count = 0
        self.path_len_diff = Counter()

        self._lock = threading.Lock()
        self.handlers = {
            STARStatEvent.UPDATE_MSG_SENT: self.on_update_msg_sent,
            STARStatEvent.LSU_MSG_SENT: self.on_lsu_sent,
            STARStatEvent.LSU_MSG_RECV: self.on_lsu_recv,
            STARStatEvent.APP_MSG_SENT: self.on_app_msg_sent,
            STARStatEvent.APP_MSG_RECV: self.on_app_msg_recv,
            STARStatEvent.LINK_UPDATED: self.on_link_updated
        }

    def push(self, event_type: STARStatEvent, data=None):
        if event_type in self.handlers.keys():
            self.handlers[event_type](data)

    def on_lsu_sent(self, data=1):
        with self._lock:
            self.msg_sent += 1
            self.lsu_sent += data

    def on_update_msg_sent(self, data):
        with self._lock:
            self.msg_sent += 1
            self.update_msg_sent += 1

    def on_lsu_recv(self, data=1):
        with self._lock:
            self.lsu_recv += data

    def on_app_msg_sent(self, data):
        with self._lock:
            self.msg_sent += 1
            self.app_sent += 1

    def on_app_msg_recv(self, data):
        with self._lock:
            self.app_recv += 1
            self.total_shortest_hop_count = data['shortest']
            self.total_hop_count = data['hop_count']
            diff = data['hop_count'] - data['shortest']
            self.path_len_diff.update({diff: 1})

    def on_link_updated(self, data):
        with self._lock:
            self.link_updated += 1

    def get_stats(self):
        with self._lock:
            data = {
                STARStatEvent.MSG_SENT: self.msg_sent,
                STARStatEvent.UPDATE_MSG_SENT: self.update_msg_sent,
                STARStatEvent.LSU_MSG_SENT: self.lsu_sent,
                STARStatEvent.LSU_MSG_RECV: self.lsu_recv,
                STARStatEvent.APP_MSG_SENT: self.app_sent,
                STARStatEvent.APP_MSG_RECV: self.app_recv,
                'TotalShortestHopCount': self.total_shortest_hop_count,
                'TotalHopCount': self.total_hop_count,
            }

            for i in range(0, 11):
                if i in self.path_len_diff.keys():
                    data[f'PathLenDiff_{i}'] = self.path_len_diff[i]
                else:
                    data[f'PathLenDiff_{i}'] = 0

            self.msg_sent = 0
            self.update_msg_sent = 0
            self.lsu_sent = 0
            self.lsu_recv = 0
            self.app_sent = 0
            self.app_recv = 0
            self.total_shortest_hop_count = 0
            self.total_hop_count = 0
            self.path_len_diff = Counter()

        return data


class MessageGenerator:
    def __init__(self, mps=1, sender_fn=None):
        """
        mps: Number of messages per second
        sender_fn: Function to be called
        """
        durations = [1, 2, 4, 8]
        self.mps = mps if mps in durations else 1
        self.sleep_time: float = 1.0 / self.mps
        self.sender_fn = sender_fn
        self.terminated = False

    def terminate(self):
        self.terminated = True

    def start(self):
        self.terminated = False

        while not self.terminated:
            self.sender_fn()
            time.sleep(self.sleep_time)



from ahc.Ahc import ComponentModel, Event
from ahc.Routing.GSR.Constants import GSR_ROUTER_NAME, GSR_COORDINATOR_NAME
from ahc.Routing.GSR.GSRExperimentDataCollector import GSRExperimentCollector


class GSRMiddlewareMessageStub(ComponentModel):
    def __init__(self, componentname, componentid):
        super(GSRMiddlewareMessageStub, self).__init__(componentname, componentid)
        pass

    def on_init(self, eventobj: Event):
        pass

    def on_message_from_bottom(self, eventobj: Event):
        message_target = eventobj.eventcontent.header.messageto.split("-")[0]
        if message_target == GSR_ROUTER_NAME or message_target == GSR_COORDINATOR_NAME:
            self.send_up(eventobj)
        else:
            print(f"VVVVVVVVVVVV******** {message_target}")

    def on_message_from_top(self, eventobj: Event):
        if self.componentinstancenumber not in GSRExperimentCollector().MESSAGE_COUNT:
            GSRExperimentCollector().MESSAGE_COUNT[self.componentinstancenumber] = 1
        else:
            GSRExperimentCollector().MESSAGE_COUNT[self.componentinstancenumber] += 1

        self.send_down(eventobj)



from enum import Enum
import threading
from ahc.Ahc import ComponentModel, Event, GenericMessageHeader, GenericMessagePayload, GenericMessage, Topology, \
  MessageDestinationIdentifiers, EventTypes
from ahc.Routing.AODV.AODVUtils import AODVMessageTypes, AODVMessageHeader
import random
# define your own message types
class NetworkLayerMessageTypes(Enum):
  NETMSG = "NETMSG"

# define your own message payload structure
class NetworkLayerMessagePayload(GenericMessagePayload):
  pass

class AODVNetworkLayerEventType(Enum):
      RREQ = "RREQ"
      RREP = "RREP"
      PROPOSE = "PROPOSE"

class AODVNetworkLayerComponent(ComponentModel):

  def on_message_from_top(self, eventobj: Event):
    #print(f"On MSFRT {self.componentname}.{self.componentinstancenumber}")
    msg = eventobj.eventcontent
    hdr = msg.header

    if hdr.messagetype == AODVMessageTypes.PROPOSE:
      if hdr.messageto in self.RoutingTable:
        pass
        #self.send_self(Event(self, AODVNetworkLayerEventType.PROPOSE),msg)
      else:
        self.send_self(Event(self, AODVNetworkLayerEventType.RREQ, msg))
    else:
      print(f"Should not be in here {self.componentname}.{self.componentinstancenumber}")

  def on_message_from_bottom(self, eventobj: Event):
    self.lock.acquire()
    #print(f"On MSFRB {self.componentname}.{self.componentinstancenumber}")
    msg = eventobj.eventcontent
    hdr = msg.header
    payload = msg.payload

    if hdr.messagetype == AODVMessageTypes.RREQ:
      hdr.hopcount += 1
      self.RoutingTable[hdr.messagefrom] = [hdr.messagefrom,hdr.interfaceid,hdr.hopcount,hdr.sequencenumber]
      #print(f"RREQ package received from MFRB")
      self.send_self(Event(self,AODVNetworkLayerEventType.RREQ,msg))
    elif hdr.messagetype == AODVMessageTypes.RREP:
      #print(f"RREP package on route came to {self.componentinstancenumber}")
      self.send_self(Event(self, AODVNetworkLayerEventType.RREP,msg)) 
    elif hdr.messagetype == AODVMessageTypes.PROPOSE:
      #print(f"PROPOSE package on route came to {self.componentinstancenumber}")
      if hdr.messageto == self.componentinstancenumber: #Data is arrived successfully to the Dest with payload
        self.send_up(Event(self,AODVNetworkLayerEventType.PROPOSE,msg))
      else: #Continue to forward
        self.send_self(Event(self,AODVMessageTypes.PROPOSE,msg))  
    self.lock.release()

  #Route Discovery: route request
  def on_rreq(self, eventobj: Event):
    self.lock.acquire()
    #print(f"On {eventobj.eventcontent.header.messagetype} {self.componentname}.{self.componentinstancenumber}")
    msg = eventobj.eventcontent
    hdr = msg.header

    neighbors = Topology().get_neighbors(self.componentinstancenumber)

    #if hdr.messagefrom not in self.RoutingTable:
      #if hdr.messagetype == AODVMessageTypes.RREQ:
      #  hdr.hopcount += 1
      #  self.RoutingTable[hdr.messagefrom] = [hdr.messagefrom,hdr.interfaceid,hdr.hopcount,hdr.sequencenumber]
    if hdr.messageto == self.componentinstancenumber:  #Prepare RREP because RREQ came to the destination.
      print(f"--- RREQ reached destination from {hdr.messagefrom} to {hdr.messageto} , msg type {hdr.messagetype} --- ")
      print(f"RoutingTable: {self.RoutingTable} and my ID: {self.componentinstancenumber}")
      self.send_self(Event(self, AODVNetworkLayerEventType.RREP, msg))
    else: #Node != DestNode so continue RREQ Broadcast
      #Broadcast
      for nexthop in neighbors:
        if nexthop != hdr.interfaceid and hdr.interfaceid != float('inf'):
          #print(f"{self.componentname}.{self.componentinstancenumber} nexthop is {nexthop}")            
          hdr.messagetype = AODVMessageTypes.RREQ
          hdr.nexthop = nexthop
          msg.payload = None
          hdr.interfaceid = self.componentinstancenumber
          self.send_down(Event(self,EventTypes.MFRT,msg))
    #else: #It is in routing table so if sequence number >= the coming package, then RREP if not so continue RREQ.
    #  print(f"Found {hdr.messageto} in Routing Table. TODO") #TODO 

    self.lock.release()

  #Route Discovery: route reply
  def on_rrep(self, eventobj: Event):
    self.lock.acquire()
    #print(f"On rrep {self.componentname}.{self.componentinstancenumber}")
  
    msg = eventobj.eventcontent
    hdr = msg.header
    payload = msg.payload

    if hdr.messagetype == AODVMessageTypes.RREQ: 
      if self.componentinstancenumber == hdr.messageto: 
        print(f"--- RREP begins from {hdr.messageto} to {hdr.messagefrom} ---")
        hopcount = 0
        destseqnumber = random.randint(0,1000)
        nexthop = hdr.interfaceid
        newhdr = AODVMessageHeader(AODVMessageTypes.RREP, hdr.messageto,hdr.messagefrom
                                ,hopcount,nexthop,self.componentinstancenumber,destseqnumber)
        newmessage = GenericMessage(newhdr, None)
        self.send_down(Event(self,EventTypes.MFRT,newmessage))
      else:
        print(f"Something is wrong on on_rrep")
    elif hdr.messagetype == AODVMessageTypes.RREP:  #RREP packages on route
      #print(f"RREP {self.componentinstancenumber} on {self.componentname} and hdr.messageto {hdr.messageto}")
      if hdr.messageto == self.componentinstancenumber: #RREP arrives at the source finally.
        #print(f"RREP arrives at the {self.componentinstancenumber} on {self.componentname}")
        self.send_up(Event(self,EventTypes.MFRB,msg))
      else: #RREP is on route
        #print(f"{self.componentinstancenumber} routing table: {self.RoutingTable}")
        if hdr.messageto in self.RoutingTable:
          #REVERSE PATH FORWARDING
          #print(f"It is in table so RREP continues")
          hdr.hopcount += 1
          nexthop = self.RoutingTable[hdr.messageto][1] #Nexthop
          self.RoutingTable[hdr.messagefrom] = [hdr.messagefrom,hdr.interfaceid,hdr.hopcount,hdr.sequencenumber]

          newhdr = AODVMessageHeader(AODVMessageTypes.RREP, hdr.messageto,hdr.messagefrom
                                ,hdr.hopcount,nexthop,self.componentinstancenumber,hdr.sequencenumber)
          newmessage = GenericMessage(newhdr, None)
          self.send_down(Event(self,EventTypes.MFRT,newmessage))
        else:
          #TODO
          print(f"Something is wrong RREP should start but {hdr.messagefrom} is not in Routing Table {self.RoutingTable}, my ID {self.componentinstancenumber}")
          print(f"Header info: {hdr}") 
      
    self.lock.release()

  def on_propose(self, eventobj: Event):
    self.lock.acquire()
    #print(f"On_Propose {self.componentname}.{self.componentinstancenumber}")
    msg = eventobj.eventcontent
    hdr = msg.header
    if hdr.messageto in self.RoutingTable: 
      #TODO
      #self.send_down(Event(self,AODVNetworkLayerEventType.PROPOSE,eventobj))
      pass
    self.lock.release()

  def __init__(self, componentname, componentinstancenumber):
    super().__init__(componentname, componentinstancenumber)
    #DestinationNode   NextNode    HopCount     SequenceNumber
    self.RoutingTable = {}
    #self.RoutingTable = { 7:[7,8,3,1] ,
    #                      6: [6,3,5,1]
    #                    }
    self.lock = threading.Lock()
    self.eventhandlers[AODVNetworkLayerEventType.RREQ] = self.on_rreq
    self.eventhandlers[AODVNetworkLayerEventType.RREP] = self.on_rrep
    self.eventhandlers[AODVNetworkLayerEventType.PROPOSE] = self.on_propose



import time
from enum import Enum
from threading import Lock
from typing import Dict, List, Tuple

from ahc.Ahc import (
    ComponentModel,
    ConnectorTypes,
    Event,
    EventTypes,
    GenericMessage,
    GenericMessageHeader,
    GenericMessagePayload,
    Topology,
)
from ahc.LinkLayers.GenericLinkLayer import LinkLayer
from ahc.Routing.AllSeeingEyeNetworkLayer import AllSeingEyeNetworkLayer

INITIAL_TIME = 20000000000
benchmark_time_lock: Lock = Lock()
benchmark_time: float = 20000000000


class ApplicationLayerMessageTypes(Enum):
    UPD = "UPDATE"
    CLR = "CLEAR"
    QRY = "QUERY"
    MSG = "MESSAGE"


class TORAHeight:
    def __init__(self, tau: float, oid: int, r: int, delta: int, i: int):
        self.tau = tau
        self.oid = oid
        self.r = r
        self.delta = delta
        self.i = i


# define your own message header structure
class ApplicationLayerMessageHeader(GenericMessageHeader):
    pass


# define your own message payload structure
class ApplicationLayerMessagePayload(GenericMessagePayload):
    pass


class ApplicationLayerQueryMessagePayload(GenericMessagePayload):
    did: int

    def __init__(self, did: int):
        self.did = did


class ApplicationLayerClearMessagePayload(GenericMessagePayload):
    did: int
    reference: Tuple[int, int, int]

    def __init__(self, did: int, reference: Tuple[int, int, int]):
        self.did = did
        self.reference = reference


class ApplicationLayerUpdateMessagePayload(GenericMessagePayload):
    did: int
    height: TORAHeight
    link_reversal: bool

    def __init__(self, did: int, height: TORAHeight, link_reversal: bool):
        self.did = did
        self.height = height
        self.link_reversal = link_reversal


class ApplicationLayerMessageMessagePayload(GenericMessagePayload):
    did: int
    message: str

    def __init__(self, did: int, message: str):
        self.did = did
        self.message = message


class RoutingTORAApplicationLayerComponent(ComponentModel):
    def __init__(self, componentname, componentinstancenumber):
        super().__init__(componentname, componentinstancenumber)
        self.neighbors = Topology().get_neighbors(componentinstancenumber)

        self.height: TORAHeight = TORAHeight(
            None, None, None, None, self.componentinstancenumber
        )

        self.last_upd: int = 0
        self.rr: bool = 0
        self.N: Dict[int, Tuple[TORAHeight, int]] = {}
        self.lock: Lock = Lock()

    def on_init(self, eventobj: Event):
        pass

    def on_message_from_bottom(self, eventobj: Event):
        self.update_time()
        with self.lock:
            try:
                applmessage = eventobj.eventcontent
                hdr = applmessage.header
                payload: GenericMessagePayload = applmessage.payload
                if hdr.messagetype == ApplicationLayerMessageTypes.QRY:
                    self.handle_qry(payload.did, hdr.messagefrom)
                elif hdr.messagetype == ApplicationLayerMessageTypes.UPD:
                    self.handle_upd(
                        payload.did,
                        hdr.messagefrom,
                        payload.height,
                        payload.link_reversal,
                    )
                elif hdr.messagetype == ApplicationLayerMessageTypes.CLR:
                    self.handle_clr(payload.did, payload.reference)
                elif hdr.messagetype == ApplicationLayerMessageTypes.MSG:
                    self.handle_msg(payload.did, payload.message)
            except AttributeError:
                print("Attribute Error")
        self.update_time()

    def handle_qry(self, did: int, fromid: int):
        downstream_links = self.get_downstream_links()

        if len(downstream_links) == 0:
            if self.rr == 0:
                self.broadcast_qry(did)
            else:
                pass
        elif self.height.delta is None:
            min_height = self.get_minimum_height_between_neighbours()
            self.height = TORAHeight(
                min_height.tau,
                min_height.oid,
                min_height.r,
                min_height.delta + 1,
                self.componentinstancenumber,
            )
            self.broadcast_upd(did, False)
        elif fromid not in self.N or (
            fromid in self.N and self.N[fromid][1] > self.last_upd
        ):
            self.broadcast_upd(did, False)
        else:
            pass

    def handle_upd(
        self, did: int, from_id: int, height: TORAHeight, link_reversal: bool
    ):
        self.set_neighbour_height(from_id, height)
        downstream_links = self.get_downstream_links()

        if link_reversal:
            
            if len(downstream_links):
                return

            upstream_links: List[Tuple[TORAHeight, int]] = list(
                self.get_upstream_links().items()
            )
            reference_level: TORAHeight = TORAHeight(-1, None, None, None, None)
            same_reference_level = True

            for t in upstream_links:
                upstream_link = t[0]
                if reference_level == (-1, None, None):
                    reference_level = upstream_link
                elif (
                    upstream_link.tau != reference_level.tau
                    or upstream_link.oid != reference_level.oid
                    or upstream_link.r != reference_level.r
                ):
                    same_reference_level = False

                if (reference_level.tau, reference_level.oid, reference_level.r) >= (
                    upstream_link.tau,
                    upstream_link.oid,
                    upstream_link.r,
                ):
                    reference_level.tau = upstream_link.tau
                    reference_level.oid = upstream_link.oid
                    reference_level.r = upstream_link.r
                    reference_level.delta = min(
                        reference_level.delta, upstream_link.delta
                    )

            if not same_reference_level:
                self.maintenance_case_2(did, reference_level)
            elif reference_level[2] == 0:
                self.maintenance_case_3(did, reference_level)
            elif self.componentinstancenumber == reference_level[1]:
                self.maintenance_case_4(did)
            else:
                self.maintenance_case_5(did)
        else:
            if self.rr == 1:
                min_height = self.get_minimum_height_between_neighbours()
                self.height = TORAHeight(
                    min_height.tau,
                    min_height.oid,
                    min_height.r,
                    min_height.delta + 1,
                    self.componentinstancenumber,
                )
                self.rr = 0
                self.broadcast_upd(did, False)
            else:
                if len(downstream_links) == 0 and self.componentinstancenumber != did:
                    self.maintenance_case_1(did)

    def handle_clr(self, did: int, reference: Tuple[int, int, int]):
        if reference == (self.height.tau, self.height.oid, self.height.r):
            self.height = TORAHeight(
                None, None, None, None, self.componentinstancenumber
            )

        for neighbour in self.N:
            if neighbour == did:
                continue
            if reference == (
                self.height.tau,
                self.height.oid,
                self.height.r,
            ) or reference == (
                self.N[neighbour][0].tau,
                self.N[neighbour][0].oid,
                self.N[neighbour][0].r,
            ):
                self.N[neighbour] = (
                    None,
                    None,
                    None,
                    None,
                    self.componentinstancenumber,
                )
        if reference == (self.height.tau, self.height.oid, self.height.r):
            self.broadcast_clr(did, reference)

    def handle_msg(self, did: int, message: str):
        if did == self.componentinstancenumber:
            print(f"{self.componentinstancenumber} says: Got a message: {message}")
        elif len(self.get_downstream_links()) == 0:
            print(
                f"{self.componentinstancenumber} says: Sorry I can't route this message :("
            )
        else:
            min_neighbour = self.get_minimum_height_between_neighbours()
            print(
                f"{self.componentinstancenumber} says: Forwarding a message: {message}"
            )
            hdr = ApplicationLayerMessageHeader(
                ApplicationLayerMessageTypes.MSG,
                self.componentinstancenumber,
                min_neighbour.i,
            )
            msg = GenericMessage(
                hdr, ApplicationLayerMessageMessagePayload(did, message)
            )
            self.send_down(Event(self, EventTypes.MFRT, msg))

    def maintenance_case_1(self, did: int):
        upstream_links = self.get_upstream_links()

        if len(upstream_links) == 0:
            self.height = TORAHeight(
                None, None, None, None, self.componentinstancenumber
            )
        else:
            self.height = TORAHeight(
                time.time(),
                self.componentinstancenumber,
                0,
                0,
                self.componentinstancenumber,
            )

        self.broadcast_upd(did, True)

    def maintenance_case_2(self, did: int, reference: TORAHeight):
        self.height = TORAHeight(
            reference.tau,
            reference.oid,
            reference.r,
            reference.delta - 1,
            self.componentinstancenumber,
        )
        self.broadcast_upd(did, True)

    def maintenance_case_3(self, did: int, reference: TORAHeight):
        self.height = TORAHeight(
            reference.tau, reference.oid, 1, 0, self.componentinstancenumber
        )
        self.broadcast_upd(did, True)

    def maintenance_case_4(self, did: int):
        self.height = TORAHeight(None, None, None, None, self.componentinstancenumber)

        for neighbour in self.N:
            if neighbour == did:
                continue
            self.N[neighbour] = (
                None,
                None,
                None,
                None,
                self.componentinstancenumber,
            )

        self.broadcast_clr(did, (self.height.tau, self.height.oid, 1))

    def maintenance_case_5(self, did: int):
        self.height = TORAHeight(
            time.time(),
            self.componentinstancenumber,
            0,
            0,
            self.componentinstancenumber,
        )
        self.broadcast_upd(did, True)

    def get_minimum_height_between_neighbours(self) -> TORAHeight:
        downstream_links = self.get_downstream_links()
        min_height = downstream_links[list(downstream_links)[0]][0]
        min_height_delta = min_height.delta

        for i in list(downstream_links):
            downstream_link = downstream_links[i]

            if min_height_delta > downstream_link[0].delta + 1:
                min_height = downstream_link[0]
                min_height_delta = downstream_link[0].delta + 1

        return min_height

    def get_downstream_links(self):
        height_delta = 100000 if self.height.delta is None else self.height.delta
        return dict(
            filter(lambda link: link[1][0].delta < height_delta, list(self.N.items()))
        )

    def get_upstream_links(self):
        height_delta = -1 if self.height.delta is None else self.height.delta
        return dict(
            filter(lambda link: link[1][0].delta >= height_delta, list(self.N.items()))
        )

    def broadcast_qry(self, did: int):
        self.rr = 1
        self.broadcast(
            ApplicationLayerQueryMessagePayload(did), ApplicationLayerMessageTypes.QRY
        )

    def broadcast_upd(self, did: int, link_reversal: bool):
        self.last_upd = time.time()
        self.broadcast(
            ApplicationLayerUpdateMessagePayload(did, self.height, link_reversal),
            ApplicationLayerMessageTypes.UPD,
        )

    def broadcast_clr(self, did: int, reference: Tuple[int, int, int]):
        self.broadcast(
            ApplicationLayerClearMessagePayload(did, reference),
            ApplicationLayerMessageTypes.CLR,
        )

    def broadcast(
        self, payload: GenericMessagePayload, t: ApplicationLayerMessageTypes
    ):
        print(f"Node-{self.componentinstancenumber} is broadcasting a {t} message")
        for destination in self.neighbors:
            hdr = ApplicationLayerMessageHeader(
                t,
                self.componentinstancenumber,
                destination,
            )
            msg = GenericMessage(hdr, payload)
            self.send_down(Event(self, EventTypes.MFRT, msg))

    def set_height(self, height: TORAHeight):
        self.height = height

        for destination_neighbour in self.neighbors:
            Topology().nodes[destination_neighbour].set_neighbour_height(
                self.componentinstancenumber, height
            )

    def set_neighbour_height(self, j: int, height: TORAHeight):
        self.N[j] = (height, time.time())

    def update_time(self):
        global benchmark_time
        with benchmark_time_lock:
            if benchmark_time < time.time() or benchmark_time == INITIAL_TIME:
                benchmark_time = time.time()


class RoutingTORAComponent(ComponentModel):
    def __init__(self, componentname, componentid):
        # SUBCOMPONENTS
        self.appllayer = RoutingTORAApplicationLayerComponent(
            "ApplicationLayer", componentid
        )
        self.netlayer = AllSeingEyeNetworkLayer("NetworkLayer", componentid)
        self.linklayer = LinkLayer("LinkLayer", componentid)
        # self.failuredetect = GenericFailureDetector("FailureDetector", componentid)

        # CONNECTIONS AMONG SUBCOMPONENTS
        self.appllayer.connect_me_to_component(ConnectorTypes.DOWN, self.netlayer)
        # self.failuredetect.connectMeToComponent(PortNames.DOWN, self.netlayer)
        self.netlayer.connect_me_to_component(ConnectorTypes.UP, self.appllayer)
        # self.netlayer.connectMeToComponent(PortNames.UP, self.failuredetect)
        self.netlayer.connect_me_to_component(ConnectorTypes.DOWN, self.linklayer)
        self.linklayer.connect_me_to_component(ConnectorTypes.UP, self.netlayer)

        # Connect the bottom component to the composite component....
        self.linklayer.connect_me_to_component(ConnectorTypes.DOWN, self)
        self.connect_me_to_component(ConnectorTypes.UP, self.linklayer)

        super().__init__(componentname, componentid)

    def on_init(self, eventobj: Event):
        pass

    def on_message_from_top(self, eventobj: Event):
        self.send_down(Event(self, EventTypes.MFRT, eventobj.eventcontent))

    def on_message_from_bottom(self, eventobj: Event):
        self.send_up(Event(self, EventTypes.MFRB, eventobj.eventcontent))

    def init_route_creation(self, did: int):
        self.appllayer.handle_qry(did, self.componentinstancenumber)

    def set_height(self, height: TORAHeight):
        self.appllayer.set_height(height)

    def send_message(self, did: int, message: str):
        self.appllayer.handle_msg(did, message)

    def set_neighbour_height(self, j: int, height: TORAHeight):
        self.appllayer.set_neighbour_height(j, height)


### Functions to get variables stored in nodes.
### Used in visiualization.


def all_edges(topo: Topology):
    edges = []
    for node in topo.nodes:
        downstream_links = topo.nodes[node].appllayer.get_downstream_links()

        for i in list(downstream_links):
            edges.append((node, i))

    return edges


def heights(topo: Topology):
    heights = []
    for node in topo.nodes:
        heights.append((node, topo.nodes[node].appllayer.height.delta))
    return heights


### Functions used in benchmark


def wait_for_action_to_complete():
    while time.time() - 0.25 < benchmark_time:
        time.sleep(0.25)
    return benchmark_time


def set_benchmark_time():
    global benchmark_time
    benchmark_time = INITIAL_TIME



from timeit import default_timer as timer

from ahc.Ahc import ComponentModel, Event, GenericMessage, GenericMessageHeader, EventTypes
from ahc.Routing.GSR.Constants import GSR_APPLICATION_NAME, GSR_COORDINATOR_NAME, ROUTING_COMPLETED_MESSAGE_TYPE, INFO_MESSAGE_TYPE


class GSRTestingApplicationComponent(ComponentModel):
    def __init__(self, componentname, componentid):
        super(GSRTestingApplicationComponent, self).__init__(componentname, componentid)
        self.start_time = None
        self.end_time = None

    def on_init(self, eventobj: Event):
        pass

    def on_message_from_bottom(self, eventobj: Event):
        event_header = eventobj.eventcontent.header
        message_to = event_header.messageto.split("-")[0]
        message_type = event_header.messagetype
        message = eventobj.eventcontent.payload
        if message_to == GSR_APPLICATION_NAME:
            if message_type == INFO_MESSAGE_TYPE:
                source = message["src"]
                content = message["content"]
                print(f"App {self.componentinstancenumber} has received {content} from {source}")
                self.send_info_message(source, "PING")
            elif message_type == ROUTING_COMPLETED_MESSAGE_TYPE:
                self.end_time = timer()
                print(f"App {self.componentinstancenumber} has received RoutingCompleted message")
                for i in range(1, 5):
                    payload = "PONG " + str(i)
                    self.send_info_message(i, payload)

    def send_info_message(self, dest, payload):
        message_header = GenericMessageHeader(
            INFO_MESSAGE_TYPE,
            GSR_APPLICATION_NAME + "-" + str(self.componentinstancenumber),
            GSR_COORDINATOR_NAME + "-" + str(self.componentinstancenumber))
        payload = {
            "src": self.componentinstancenumber,
            "dest": dest,
            "content": payload
        }
        message = GenericMessage(message_header, payload)
        event = Event(self, EventTypes.MFRT, message)
        self.send_down(event)

    def get_duration(self):
        return self.end_time - self.start_time



import networkx as nx
from ahc.Ahc import GenericMessage, GenericMessageHeader, ComponentRegistry, Topology

def messageParser(self, eventobj, destination = ""):
    messageTo = eventobj.eventcontent.header.messageto
    if destination != "":
        messageTo = destination
    messagePayload = eventobj.eventcontent.payload
    messageFrom = eventobj.eventcontent.header.messagefrom
    #print(f"{self.componentname}-{self.componentid} got a message from {messageFrom}. \n Message is {messagePayload}\n")
    nextHop=eventobj.eventcontent.header.nexthop

    if eventobj.eventcontent.header.messagetype == "ROUTESEARCH":
        nextHop = None
    messageHeader = GenericMessageHeader(eventobj.eventcontent.header.messagetype, eventobj.eventcontent.header.messagefrom, messageTo, interfaceid = eventobj.eventcontent.header.interfaceid, nexthop=nextHop, sequencenumber = eventobj.eventcontent.header.sequencenumber)
    message = GenericMessage(messageHeader, messagePayload)
    
    return message
           
def buildRoutingTable(sourceNode, target):
    myNode = ComponentRegistry().get_component_by_key("FP", sourceNode)
    myNode.build_routing_table(target)

def benchmarkTest(nodeCount):
    myNode = ComponentRegistry().get_component_by_key("FP", 0)
    myNode.build_routing_table(nodeCount, 1, nodeCount)

def findStrongConnectedLinksForSingleNode(labels, threshold, nodeCount):

    for nodeToBeCalculated in range (nodeCount):
        
        strongLinks = dict()
        allNeighbors = dict()
        
        for (new_key, new_value) in labels.items():
        
            if new_value >= threshold:
                if (str(new_key[0]) == str(nodeToBeCalculated)):
                    strongLinks[new_key[1]] = new_value
                elif (str(new_key[1]) == str(nodeToBeCalculated)):
                    strongLinks[new_key[0]] = new_value
            elif (str(new_key[0]) == str(nodeToBeCalculated)):
                    allNeighbors[new_key[1]] = new_value
            elif (str(new_key[1]) == str(nodeToBeCalculated)):
                allNeighbors[new_key[0]] = new_value

        nodeToEditSST = ComponentRegistry().get_component_by_key("DRP",nodeToBeCalculated)
        
        for (key) in strongLinks.keys():
            nodeToEditSST.editSignalStabilityTable(key, "SC")
        for (key) in allNeighbors.keys():
            nodeToEditSST.editSignalStabilityTable(key, "WC")
    
    return
    
def printSSTInfo(nodeCount):
    for x in range(nodeCount):
        print("--------------------")
        print(f"Information of {x}:\n")
        nodeToEditSST = ComponentRegistry().get_component_by_key("DRP",x)
        print("Stability Table:")
        nodeToEditSST.printSignalStabilityTable()
        print("Routing Table:")
        nodeToEditSST = ComponentRegistry().get_component_by_key("FP",x)
        print(nodeToEditSST.routingTable)
        
def resetRoutingState(nodeCount):
    for x in range (0, nodeCount):
        nodeToEdit = ComponentRegistry().get_component_by_key("DRP",x)
        nodeToEditTwo = ComponentRegistry().get_component_by_key("SSBRNode",x)
        nodeToEditThree = ComponentRegistry().get_component_by_key("FP",x)
        nodeToEdit.routingTableFlag = False
        nodeToEditTwo.messageFrom = -1
        nodeToEditThree.node = 0
        nodeToEditThree.nodeCount = 0

def constructStrongRoute(graph, source, target):
    paths = nx.all_simple_paths(graph, source, target)
    sortedPath = sorted(paths, key=len)
    for i in sortedPath:
        desiredPath = True
        for index, item in enumerate(i):
            if len(i) > index+1:
                nodeToBeInvestigated = ComponentRegistry().get_component_by_key("DRP",i[index+1])
                if nodeToBeInvestigated.getSignalStabilityTable()[i[index]] == "WC":
                    desiredPath = False
        if desiredPath:
            return i
    return []
  
def SSBRRouteSearchMessage(self, target):
    message_payload = []
    messageFrom = str(self.componentname) + "-" + str(self.componentinstancenumber)
    messageTo = "FP" + "-" + str(target)
    message_header = GenericMessageHeader("ROUTESEARCH", messageFrom ,messageTo ,sequencenumber=0)
    message = GenericMessage(message_header, message_payload)

    return message

def SSBRRouteReplyMessage(self, eventobj):
    
    messagePayload = []
    messageFrom = eventobj.eventcontent.header.messagefrom
    messageTo = eventobj.eventcontent.header.messageto
    sequenceNumber = 0
    interfaceid = eventobj.eventcontent.header.interfaceid

    if int(self.componentinstancenumber) == int(interfaceid.split("-")[0]):
        nextHop = int(interfaceid.split("-")[1])
    else:
        nextHop = int(interfaceid.split("-")[0])
    
    messageHeader = GenericMessageHeader("ROUTEREPLY", messageTo, messageFrom, nextHop, interfaceid, sequenceNumber)
    message = GenericMessage(messageHeader, messagePayload)

    return message

def sendMessageToOtherNode(self, eventobj, nodeid):

    messageFrom = eventobj.eventcontent.header.messagefrom
    messageTo = eventobj.eventcontent.header.messageto
    if int(self.componentinstancenumber) > int(nodeid):
        interfaceid = str(nodeid)+"-"+str(self.componentinstancenumber)
    else:
        interfaceid=str(self.componentinstancenumber)+"-"+str(nodeid)

    sequenceNumber = int(eventobj.eventcontent.header.sequencenumber) + 1
    nextHop = nodeid
    messageType = eventobj.eventcontent.header.messagetype    
    messagePayload = eventobj.eventcontent.payload

    if eventobj.eventcontent.header.messagetype == "ROUTEREPLY":
        if int(eventobj.eventcontent.header.sequencenumber) == 0:
            messagePayload = []
        if self.componentinstancenumber not in messagePayload: messagePayload.append(self.componentinstancenumber)
        
    messageHeader = GenericMessageHeader(messageType, messageFrom, messageTo, nextHop, interfaceid, sequenceNumber)    
    message = GenericMessage(messageHeader, messagePayload)
    #print(messageHeader)
    #print(f"{nextHop} got a {eventobj.eventcontent.header.messagetype} message from {self.componentid}.\n")
    return message

def SSBRRouteCompletedMessage(self, eventobj):
    
    messagePayload = eventobj.eventcontent.payload
    messageFrom = eventobj.eventcontent.header.messagefrom
    messageTo = eventobj.eventcontent.header.messageto
    sequenceNumber = eventobj.eventcontent.header.sequencenumber
    interfaceid = eventobj.eventcontent.header.interfaceid
    nextHop = eventobj.eventcontent.header.nexthop
   
    messageHeader = GenericMessageHeader("ROUTECOMPLETED", messageTo, messageFrom, nextHop, interfaceid, sequenceNumber)
    message = GenericMessage(messageHeader, messagePayload)

    return message

def SSBRUnicastMessage(self, destination, message=""):
    messagePayload = message
    if (messagePayload == ""):
        messagePayload = input("Enter payload of message...\n")
    messageFrom = str(self.componentname) + "-" + str(self.componentid)
    messageTo = str(self.componentname) + "-" + str(destination)
    sequenceNumber = 0
   
    messageHeader = GenericMessageHeader("UNICASTDATA", messageFrom, messageTo, sequencenumber=sequenceNumber)
    message = GenericMessage(messageHeader, messagePayload)

    return message

def SSBRUnicastMessageFPParser(self, eventobj):
    target = eventobj.eventcontent.header.messageto.split("-")[1]
    nextHop = self.routingTable.get(str(target))
    messageHeader = GenericMessageHeader(eventobj.eventcontent.header.messagetype, eventobj.eventcontent.header.messagefrom, eventobj.eventcontent.header.messageto, nextHop, sequencenumber=eventobj.eventcontent.header.sequencenumber)
    message = GenericMessage(messageHeader, eventobj.eventcontent.payload)

    return message

    


import queue
import random
from enum import Enum
from datetime import datetime as dt

from ahc.LinkLayers.GenericLinkLayer import LinkLayer
from ahc.Routing.AllSeeingEyeNetworkLayer import AllSeingEyeNetworkLayer
from ahc.Ahc import (ComponentModel, Event, ConnectorTypes, ComponentRegistry, GenericMessagePayload, GenericMessageHeader,
                 GenericMessage, EventTypes)

registry = ComponentRegistry()

# define your own message types
class ApplicationLayerMessageType(Enum):
    BASIC = "basic"
    CONTROL = "control"

class AHCNodeSimulationStatus(Enum):
    ACTIVE = "active"
    PASSIVE = "passive"
    OUT_OF_CLOCK = "ooc"

# define your own message header structure
class ApplicationLayerMessageHeader(GenericMessageHeader):
    pass

# define your own message payload structure
class ApplicationLayerMessagePayload(GenericMessagePayload):
    pass

class ApplicationLayerComponent(ComponentModel):
    def __init__(self, componentname, componentinstancenumber, context):
        super().__init__(componentname, componentinstancenumber, context=context)

        self.context = context
        # self.eventhandlers[ApplicationLayerMessageType.BASIC] = self.on_basic_message
        # self.eventhandlers[ApplicationLayerMessageType.CONTROL] = self.on_control_message

        self.basic_message_queue = queue.Queue(maxsize=-1)
        self.control_message_queue = queue.Queue(maxsize=-1)
        self.simulation_state = AHCNodeSimulationStatus.PASSIVE

        self.sleep_ms_per_tick = context["ms_per_tick"]
        self.simulation_ticks_total = context["simulation_ticks"]
        self.alive_for_next_ticks = context["initial_liveness"][componentinstancenumber]
        self.communication_on_active_prob = context["communication_on_active_prob"]
        self.min_activeness_after_receive = context["min_activeness_after_receive"]
        self.max_activeness_after_receive = context["max_activeness_after_receive"]
        self.package_process_per_tick = context["node_package_process_per_tick"]
        self.die_passiveness_threshold = context["passiveness_death_thresh"]

        if context["hard_stop_on_tick"] is None:
            self.hard_stop_on_tick = None
        else:
            self.hard_stop_on_tick = context["hard_stop_on_tick"][self.componentinstancenumber]

        self.friend_ids = [i for i in context["network"].G.nodes() if i != componentinstancenumber]

        self.__tick_n = 0
        self._passive_counter = 0

        if self.alive_for_next_ticks > 0:
            self.simulation_state = AHCNodeSimulationStatus.ACTIVE

    def prepare_application_layer_message(self, message_type: ApplicationLayerMessageType, destination_node_id: int, payload: object) -> GenericMessage:
        hdr = ApplicationLayerMessageHeader(message_type, self.componentinstancenumber, destination_node_id)
        payload = ApplicationLayerMessagePayload(payload)
        
        return GenericMessage(hdr, payload)

    def send_random_basic_message(self, to: int) -> None:
        self.send_down(Event(self, EventTypes.MFRT, self.prepare_application_layer_message(ApplicationLayerMessageType.BASIC, to, str(dt.now().timestamp()))))

    def on_init(self, eventobj: Event):
        # print(f"Initializing {self.componentname}.{self.componentinstancenumber}")
        pass

    def on_message_from_bottom(self, eventobj: Event):
        applmessage = eventobj.eventcontent
        hdr = applmessage.header

        # print(f"Node-{self.componentinstancenumber}: Node-{hdr.messagefrom} has sent {hdr.messagetype} message (payload: {applmessage.payload})")

        if hdr.messagetype == ApplicationLayerMessageType.BASIC:
            self.basic_message_queue.put_nowait(applmessage)
        elif hdr.messagetype == ApplicationLayerMessageType.CONTROL:
            self.control_message_queue.put_nowait(applmessage)

    def simulation_tick(self):
        next_state = None
        got_packages_from = None
        to_friend = None

        if self.__tick_n >= self.simulation_ticks_total:
            next_state = AHCNodeSimulationStatus.OUT_OF_CLOCK
        elif self._passive_counter >= self.die_passiveness_threshold:
            next_state = AHCNodeSimulationStatus.OUT_OF_CLOCK
        elif self.__tick_n >= self.hard_stop_on_tick:
            next_state = AHCNodeSimulationStatus.OUT_OF_CLOCK
            print(f"   ==> N-{self.componentinstancenumber}: HARD STOP")
        else:
            if self.simulation_state == AHCNodeSimulationStatus.OUT_OF_CLOCK:
                next_state = AHCNodeSimulationStatus.OUT_OF_CLOCK
            elif self.simulation_state == AHCNodeSimulationStatus.PASSIVE:
                if self.basic_message_queue.empty():
                    # no incoming package, still passive.
                    next_state = AHCNodeSimulationStatus.PASSIVE
                else:
                    got_packages_from = []

                    for _ in range(self.package_process_per_tick):
                        try:
                            package = self.basic_message_queue.get_nowait()
                            # print(f"+P+ N-{self.componentinstancenumber} <==BASIC== N-{package.header.messagefrom} ({package.payload.messagepayload})")
                            got_packages_from.append(package.header.messagefrom)
                        except queue.Empty:
                            break

                    self.alive_for_next_ticks = random.randint(self.min_activeness_after_receive, self.max_activeness_after_receive)
                    next_state = AHCNodeSimulationStatus.ACTIVE
            elif self.simulation_state == AHCNodeSimulationStatus.ACTIVE:
                got_packages_from = []

                for _ in range(self.package_process_per_tick):
                    try:
                        package = self.basic_message_queue.get_nowait()
                        # print(f"+A+ N-{self.componentinstancenumber} <==BASIC== N-{package.header.messagefrom} ({package.payload.messagepayload})")
                        got_packages_from.append(package.header.messagefrom)
                    except queue.Empty:
                        break

                if random.random() <= self.communication_on_active_prob:
                    # send package to a random friend..
                    to_friend = random.choice(self.friend_ids)
                    self.send_random_basic_message(to_friend)

                self.alive_for_next_ticks -= 1

                if self.alive_for_next_ticks == 0:
                    if len(got_packages_from) > 0:
                        # got a package, this means immeiate activeness from passive!
                        next_state = AHCNodeSimulationStatus.ACTIVE
                        self.alive_for_next_ticks = random.randint(self.min_activeness_after_receive, self.max_activeness_after_receive)
                    else:
                        next_state = AHCNodeSimulationStatus.PASSIVE
                else:
                    next_state = AHCNodeSimulationStatus.ACTIVE

        assert next_state is not None

        # ST: state
        # NS: next state
        # GPF: got packages from
        # SPF: sent package to friend
        # ANT: alive next ticks
        # P2P: packages to process
        # print(f"   ==> N-{self.componentinstancenumber}: ST: {self.simulation_state}, NS: {next_state}, GPF: {got_packages_from}, SPF: {to_friend}, ANT: {self.alive_for_next_ticks}, P2P: {self.basic_message_queue.qsize()}")

        # time.sleep(self.sleep_ms_per_tick / 1000)
        self.__tick_n += 1
        self.simulation_state = next_state

        if self.simulation_state == AHCNodeSimulationStatus.PASSIVE:
            self._passive_counter += 1
        elif self.simulation_state == AHCNodeSimulationStatus.ACTIVE:
            self._passive_counter = 0

        return next_state, to_friend

class AdHocNode(ComponentModel):
    def __init__(self, componentname, componentid, context):
        self.context = context
        # SUBCOMPONENTS
        self.appllayer = ApplicationLayerComponent("ApplicationLayer", componentid, context=self.context)
        self.netlayer = AllSeingEyeNetworkLayer("NetworkLayer", componentid)
        self.linklayer = LinkLayer("LinkLayer", componentid)
        # self.failuredetect = GenericFailureDetector("FailureDetector", componentid)

        # CONNECTIONS AMONG SUBCOMPONENTS
        self.appllayer.connect_me_to_component(ConnectorTypes.DOWN, self.netlayer)
        # self.failuredetect.connectMeToComponent(PortNames.DOWN, self.netlayer)
        self.netlayer.connect_me_to_component(ConnectorTypes.UP, self.appllayer)
        # self.netlayer.connectMeToComponent(PortNames.UP, self.failuredetect)
        self.netlayer.connect_me_to_component(ConnectorTypes.DOWN, self.linklayer)
        self.linklayer.connect_me_to_component(ConnectorTypes.UP, self.netlayer)

        # Connect the bottom component to the composite component....
        self.linklayer.connect_me_to_component(ConnectorTypes.DOWN, self)
        self.connect_me_to_component(ConnectorTypes.UP, self.linklayer)

        super().__init__(componentname, componentid, context=self.context)

    def on_init(self, eventobj: Event):
        # print(f"Initializing {self.componentname}.{self.componentinstancenumber}")
        pass

    def on_message_from_top(self, eventobj: Event):
        self.send_down(Event(self, EventTypes.MFRT, eventobj.eventcontent))

    def on_message_from_bottom(self, eventobj: Event):
        self.send_up(Event(self, EventTypes.MFRB, eventobj.eventcontent))

    def simulation_tick(self):
        return self.appllayer.simulation_tick()

    @property
    def waiting_packages_on_queue(self):
        if self.appllayer.simulation_state == AHCNodeSimulationStatus.OUT_OF_CLOCK:
            return 0

        return self.appllayer.basic_message_queue.qsize()


#!/usr/bin/env python

__author__ = "One solo developer"
__authors__ = ["Mahmoud Alasmar"]
__contact__ = "mahmoud.asmar@metu.edu.tr"
__date__ = "2021/05/26"
__license__ = "GPLv3"
__maintainer__ = "developer"
__status__ = "Production"
__version__ = "0.0.1"


import random
from enum import Enum
from ahc.Ahc import ComponentModel, Event, GenericMessagePayload, GenericMessageHeader, GenericMessage, EventTypes
from ahc.Ahc import Topology
topo = Topology()

class ApplicationLayerMessageTypes(Enum):
  DISCOVER = "DISCOVER"
  RETURN = "RETURN"
  VISITED = "VISITED"
  ACK = "ACK"

# define your own message header structure
class ApplicationLayerMessageHeader(GenericMessageHeader):
  pass

# define your own message payload structure
class ApplicationLayerMessagePayload(GenericMessagePayload):
  pass

class WaveAwerbuchComponent(ComponentModel):
  def on_init(self, eventobj: Event):
    print(f"Initializing {self.componentname}.{self.componentinstancenumber}")
    neighbour_list = topo.get_neighbors(self.componentinstancenumber)
    self.NeighbourList = neighbour_list
    self.Unvisited = neighbour_list.copy()
    self.father = self.componentinstancenumber
    self.flag = {}
    for i in self.NeighbourList:
      self.flag[i] = 0

    self.numMesg = 0

    if self.componentinstancenumber == 0:
      destination = self.componentinstancenumber
      hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.DISCOVER, self.componentinstancenumber,
                                          destination)
      payload = ApplicationLayerMessagePayload("23")
      proposalmessage = GenericMessage(hdr, payload)
      self.send_self(Event(self, "discover", proposalmessage))
    else:
      pass

  def on_message_from_bottom(self, eventobj: Event):
    try:
      applmessage = eventobj.eventcontent
      hdr = applmessage.header
      if hdr.messagetype == ApplicationLayerMessageTypes.DISCOVER:
        print(f"Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} has sent {hdr.messagetype} message")
        self.send_self(Event(self, "discover", applmessage))
      elif hdr.messagetype == ApplicationLayerMessageTypes.VISITED:
        print(f"Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} has sent {hdr.messagetype} message")
        self.send_self(Event(self, "visited", applmessage))
      elif hdr.messagetype == ApplicationLayerMessageTypes.RETURN:
        print(f"Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} has sent {hdr.messagetype} message")
        self.send_self(Event(self, "return", applmessage))
      elif hdr.messagetype == ApplicationLayerMessageTypes.ACK:
        print(f"Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} has sent {hdr.messagetype} message")
        self.send_self(Event(self, "ack", applmessage))
    except AttributeError:
      print("Attribute Error")

  def on_discover(self, eventobj: Event):
    applmessage = eventobj.eventcontent
    hdr = applmessage.header
    self.father = hdr.messagefrom
    for i in self.NeighbourList:
      if i != self.father:
        self.flag[i] = 1
        hdr_new = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.VISITED, self.componentinstancenumber,i)
        payload = ApplicationLayerMessagePayload("23")  # payload is redundant
        proposalmessage = GenericMessage(hdr_new, payload)
        self.send_down(Event(self, EventTypes.MFRT, proposalmessage))
        self.numMesg+=1
      else:
        pass

    if len(self.NeighbourList) == 1 and self.NeighbourList[0] == self.father:
      hdr_new = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.RETURN, self.componentinstancenumber,
                                              self.NeighbourList[0])
      payload = ApplicationLayerMessagePayload("23")
      proposalmessage = GenericMessage(hdr_new, payload)
      self.send_down(Event(self, EventTypes.MFRT, proposalmessage))
      self.numMesg += 1



  def on_return(self, eventobj: Event):

    if not self.Unvisited:
      if self.father != self.componentinstancenumber:
        hdr_new = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.RETURN, self.componentinstancenumber,
                                                self.father)
        payload = ApplicationLayerMessagePayload("23")
        proposalmessage = GenericMessage(hdr_new, payload)
        self.send_down(Event(self, EventTypes.MFRT, proposalmessage))
        self.numMesg += 1
        print(f"I am Node-{self.componentinstancenumber} local number messages sent is {self.numMesg}")
      else:
        print(f"I am Node-{self.componentinstancenumber} local number messages sent is {self.numMesg}")
        print(f"I am Node-{self.componentinstancenumber}, algorithm is finished ")

    else:
      destination = random.choice(self.Unvisited)
      hdr_new = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.DISCOVER, self.componentinstancenumber,
                                              destination)
      payload = ApplicationLayerMessagePayload("23")
      proposalmessage = GenericMessage(hdr_new, payload)
      self.send_down(Event(self, EventTypes.MFRT, proposalmessage))
      self.numMesg += 1
      self.Unvisited.remove(destination)
      print(f"I am Node-{self.componentinstancenumber} sending DISCOVER to {destination}")

  def on_visited(self, eventobj: Event):
    applmessage = eventobj.eventcontent
    hdr = applmessage.header
    k = hdr.messagefrom
    self.Unvisited.remove(k)
    hdr_new = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.ACK, self.componentinstancenumber, k)
    payload = ApplicationLayerMessagePayload("23")
    proposalmessage = GenericMessage(hdr_new, payload)
    self.send_down(Event(self, EventTypes.MFRT, proposalmessage))
    self.numMesg += 1


  def on_ack(self, eventobj: Event):
    applmessage = eventobj.eventcontent
    hdr = applmessage.header
    k = hdr.messagefrom
    self.flag[k] = 0
    cnt = 0
    for i in self.NeighbourList:
      cnt+=self.flag[i]

    if cnt == 0:
      self.send_self(Event(self, "return", applmessage))
    else:
      pass




  def __init__(self, componentname, componentinstancenumber):
    super().__init__(componentname, componentinstancenumber)
    self.eventhandlers["discover"] = self.on_discover
    self.eventhandlers["return"] = self.on_return
    self.eventhandlers["visited"] = self.on_visited
    self.eventhandlers["ack"] = self.on_ack







import os
import pickle
import seaborn as sns
import matplotlib.pyplot as plt

metrics = {
    "DS": {
        "erg": [],
        "star": [],
        "grid": []
    },
    "SF": {
        "erg": [],
        "star": [],
        "grid": []
    }
}

latency = {
    "DS": {
        "erg": [],
        "star": [],
        "grid": []
    },
    "SF": {
        "erg": [],
        "star": [],
        "grid": []
    }
}

lnr_distr = {
    "DS": {
        "erg": None,
        "star": None,
        "grid": None
    },
    "SF": {
        "erg": None,
        "star": None,
        "grid": None
    }
}

snr_distrib = {
    "DS": {
        "erg": {},
        "star": {},
        "grid": {}
    },
    "SF": {
        "erg": {},
        "star": {},
        "grid": {}
    }
}


for fpath in os.listdir("bench_dump/simdump/"):
    if fpath.endswith(".pkl"):
        algo, _, topo, nodes, _ = fpath.split("_")
        nodes = int(nodes)

        # print(f"++ {algo} {topo} {nodes}")

        with open(f"bench_dump/simdump/{fpath}", "rb") as fp:
            data = pickle.load(fp)

        cpr = list(data["stats"]["df"]["control_total_cumulative_ratio"])[-1]
        cppn = sum(list(data["stats"]["df"]["control_packets_sent"])) / nodes

        if algo == "SF":
            wpr = list(data["stats"]["df"]["wave_total_cumulative_ratio"])[-1]
            wppn = sum(list(data["stats"]["df"]["wave_packets_sent"])) / nodes

            metrics["SF"][topo].append((nodes, {
                "cpr": cpr,
                "wpr": wpr,
                "cppn": cppn,
                "wppn": wppn,
                "cwpr": cpr + wpr
            }))
        elif algo == "DS":
            metrics["DS"][topo].append((nodes, {
                "cpr": cpr,
                "cppn": cppn,
            }))
        else:
            raise RuntimeError(algo)

for fpath in os.listdir("bench_dump/"):
    if fpath.endswith(".out"):
        algo, topo, nodes, _ = fpath.split("_")
        nodes = int(nodes)

        if topo == "grid":
            nodes = nodes ** 2
        elif topo == "star":
            nodes = nodes + 1
        elif topo == "erg":
            pass
        else:
            raise RuntimeError(topo)

        # print(f"## {algo} {topo} {nodes} ({fpath})")

        try:
            with open(f"bench_dump/{fpath}", "r") as fp:
                lines = fp.readlines()
        except UnicodeDecodeError as e:
            print(f"[!!!] UnicodeDecodeError ({fpath}): {e}")
            continue

        if algo == "SF":
            reason = lines[-1].split(" ")[0]

            if reason not in ["wave", "forced"]:
                raise RuntimeError(reason)

            if reason == "forced":
                latency["SF"][topo].append((False, ))

                if nodes not in snr_distrib["SF"][topo]:
                    snr_distrib["SF"][topo][nodes] = []

                snr_distrib["SF"][topo][nodes].append(0)
            else:
                reason, _, _, _, diff = lines[-1].split()

                assert diff[0] == "[" and diff[-1] == "]"

                if diff[1:-1] != "None":
                    dif = int(diff[1:-1])
                else:
                    dif = 0

                latency["SF"][topo].append((True, dif, dif/nodes))

                if nodes not in snr_distrib["SF"][topo]:
                    snr_distrib["SF"][topo][nodes] = []

                snr_distrib["SF"][topo][nodes].append(1)
        elif algo == "DS":
            reason = lines[-1].split(" ")[0]

            if reason not in ["root", "forced"]:
                raise RuntimeError(reason)

            if reason == "forced":
                latency["DS"][topo].append((False, ))

                if nodes not in snr_distrib["DS"][topo]:
                    snr_distrib["DS"][topo][nodes] = []

                snr_distrib["DS"][topo][nodes].append(0)
            else:
                reason, _, _, diff = lines[-1].split()

                assert diff[0] == "[" and diff[-1] == "]"

                if diff[1:-1] != "None":
                    dif = int(diff[1:-1])
                else:
                    dif = 0

                latency["DS"][topo].append((True, dif, dif/nodes))

                if nodes not in snr_distrib["DS"][topo]:
                    snr_distrib["DS"][topo][nodes] = []

                snr_distrib["DS"][topo][nodes].append(1)
        else:
            raise RuntimeError(algo)

for algo in lnr_distr:
    for topo in lnr_distr[algo]:
        lnr_distr[algo][topo] = [x[-1] for x in latency[algo][topo] if x[0] and x[-1] > 0]

# print(json.dumps({
#     "metrics": metrics,
#     "latency": latency,
#     "lnr": lnr_distr
# }, indent=4))

fig, axes = plt.subplots(2, 3)
fig.set_figheight(10)
fig.set_figwidth(25)

axes[0][0].set_title("DS/SF Grid Topology LNR Distribution")
axes[0][1].set_title("DS/SF Star Topology LNR Distribution")
axes[0][2].set_title("DS/SF ERG Topology LNR Distribution")

axes[0][0].set_xlabel("Latency/Node Ratio (ticks/node)")
axes[0][1].set_xlabel("Latency/Node Ratio (ticks/node)")
axes[0][2].set_xlabel("Latency/Node Ratio (ticks/node)")

sns.histplot(data={"Dijkstra-Scholten": lnr_distr["DS"]["grid"], "Shavit-Francez": lnr_distr["SF"]["grid"]}, kde=True, ax=axes[0][0], log_scale=True)
sns.histplot(data={"Dijkstra-Scholten": lnr_distr["DS"]["star"], "Shavit-Francez": lnr_distr["SF"]["star"]}, kde=True, ax=axes[0][1], log_scale=True)
sns.histplot(data={"Dijkstra-Scholten": lnr_distr["DS"]["erg"], "Shavit-Francez": lnr_distr["SF"]["erg"]}, kde=True, ax=axes[0][2], log_scale=True)

axes[1][0].set_title("DS/SF Grid SNR Topology Plot")
axes[1][1].set_title("DS/SF Star SNR Topology Plot")
axes[1][2].set_title("DS/SF ERG SNR Topology Plot")

axes[1][0].set_xlabel("Node Count")
axes[1][1].set_xlabel("Node Count")
axes[1][2].set_xlabel("Node Count")

axes[1][0].set_ylabel("Successive Simulations (%)")
axes[1][1].set_ylabel("Successive Simulations (%)")
axes[1][2].set_ylabel("Successive Simulations (%)")

topos = [
    "grid",
    "star",
    "erg"
]

for i in range(3):
    sns.lineplot(y=[100 * sum(snr_distrib["DS"][topos[i]][node_count]) / len(snr_distrib["DS"][topos[i]][node_count]) for node_count in snr_distrib["DS"][topos[i]]], x=[node_count for node_count in snr_distrib["DS"][topos[i]]], ax=axes[1][i], legend='brief', label="Dijkstra-Scholten", alpha=0.7)  
    sns.lineplot(y=[100 * sum(snr_distrib["SF"][topos[i]][node_count]) / len(snr_distrib["SF"][topos[i]][node_count]) for node_count in snr_distrib["SF"][topos[i]]], x=[node_count for node_count in snr_distrib["SF"][topos[i]]], ax=axes[1][i], legend='brief', label="Shavit-Francez", alpha=0.7)  

sns.despine()
plt.savefig("lnr_snr.png", dpi=200)

fig, axes = plt.subplots(2, 3)
fig.set_figheight(10)
fig.set_figwidth(25)

axes[0][0].set_title("DS/SF Grid Control Package Ratio Plot")
axes[0][1].set_title("DS/SF Star Control Package Ratio Plot")
axes[0][2].set_title("DS/SF ERG Control Package Ratio Plot")

axes[0][0].set_xlabel("Node Count")
axes[0][1].set_xlabel("Node Count")
axes[0][2].set_xlabel("Node Count")

axes[0][0].set_ylabel("Average CPR (%)")
axes[0][1].set_ylabel("Average CPR (%)")
axes[0][2].set_ylabel("Average CPR (%)")

# axes[0][0].set_yscale("log")
# axes[0][1].set_yscale("log")
# axes[0][2].set_yscale("log")

for i in range(3):
    _top = topos[i]
    cprs = {}
    c_wprs = {}
    c_prs = {}
    wprs = {}

    for x in metrics["DS"][_top]:
        node_count = x[0]
        cpr = x[1]["cpr"]

        if node_count not in cprs:
            cprs[node_count] = []

        cprs[node_count].append(cpr)

    for x in metrics["SF"][_top]:
        node_count = x[0]

        cw_pr = x[1]["cwpr"]
        c_pr = x[1]["cpr"]
        wpr = x[1]["wpr"]

        if node_count not in c_wprs:
            c_wprs[node_count] = []

        if node_count not in c_prs:
            c_prs[node_count] = []

        if node_count not in wprs:
            wprs[node_count] = []

        c_wprs[node_count].append(cw_pr)
        c_prs[node_count].append(c_pr)
        wprs[node_count].append(wpr)

    sns.lineplot(y=[sum(cprs[node_count]) / len(cprs[node_count]) for node_count in cprs], x=[node_count for node_count in cprs], ax=axes[0][i], legend='brief', label="Dijkstra-Scholten (CPR)")  
    sns.lineplot(y=[sum(c_prs[node_count]) / len(c_prs[node_count]) for node_count in c_prs], x=[node_count for node_count in c_prs], ax=axes[0][i], legend='brief', label="Shavit-Francez (CPR)")  
    sns.lineplot(y=[sum(wprs[node_count]) / len(wprs[node_count]) for node_count in wprs], x=[node_count for node_count in wprs], ax=axes[0][i], legend='brief', label="Shavit-Francez (WPR)")  
    sns.lineplot(y=[sum(c_wprs[node_count]) / len(c_wprs[node_count]) for node_count in c_wprs], x=[node_count for node_count in c_wprs], ax=axes[0][i], legend='brief', label="Shavit-Francez (CWPR)")  

axes[1][0].set_title("Dijkstra-Scholten Average CPPN Ratio Plot")
axes[1][1].set_title("Shavit-Francez Average CPPN Ratio Plot")
axes[1][2].set_title("Shavit-Francez Average WPPN Ratio Plot")

axes[1][0].set_xlabel("Node Count")
axes[1][1].set_xlabel("Node Count")
axes[1][2].set_xlabel("Node Count")

axes[1][0].set_ylabel("Average CPPN (control packets/node)")
axes[1][1].set_ylabel("Average CPPN (control packets/node)")
axes[1][2].set_ylabel("Average WPPN (wave packets/node)")

def get_average_cppn(algo, topo):
    cppns = {}

    for x in metrics[algo][topo]:
        node_count = x[0]
        cppn = x[1]["cppn"]

        if node_count not in cppns:
            cppns[node_count] = []

        cppns[node_count].append(cppn)

    return ([sum(cppns[node_count]) / len(cppns[node_count]) for node_count in cppns], [node_count for node_count in cppns])

def get_average_wppn(topo):
    wppns = {}

    for x in metrics["SF"][topo]:
        node_count = x[0]
        wppn = x[1]["wppn"]

        if node_count not in wppns:
            wppns[node_count] = []

        wppns[node_count].append(wppn)

    return ([sum(wppns[node_count]) / len(wppns[node_count]) for node_count in wppns], [node_count for node_count in wppns])

ds_grid_cppns = get_average_cppn("DS", "grid")
ds_star_cppns = get_average_cppn("DS", "star")
ds_erg_cppns = get_average_cppn("DS", "erg")

sf_grid_cppns = get_average_cppn("SF", "grid")
sf_star_cppns = get_average_cppn("SF", "star")
sf_erg_cppns = get_average_cppn("SF", "erg")

sf_grid_wppns = get_average_wppn("grid")
sf_star_wppns = get_average_wppn("star")
sf_erg_wppns = get_average_wppn("erg")

sns.lineplot(y=ds_grid_cppns[0], x=ds_grid_cppns[1], ax=axes[1][0], legend='brief', label="Grid")  
sns.lineplot(y=ds_star_cppns[0], x=ds_star_cppns[1], ax=axes[1][0], legend='brief', label="Star")  
sns.lineplot(y=ds_erg_cppns[0], x=ds_erg_cppns[1], ax=axes[1][0], legend='brief', label="ERG")  

sns.lineplot(y=sf_grid_cppns[0], x=sf_grid_cppns[1], ax=axes[1][1], legend='brief', label="Grid")  
sns.lineplot(y=sf_star_cppns[0], x=sf_star_cppns[1], ax=axes[1][1], legend='brief', label="Star")  
sns.lineplot(y=sf_erg_cppns[0], x=sf_erg_cppns[1], ax=axes[1][1], legend='brief', label="ERG")  

sns.lineplot(y=sf_grid_wppns[0], x=sf_grid_wppns[1], ax=axes[1][2], legend='brief', label="Grid")  
sns.lineplot(y=sf_star_wppns[0], x=sf_star_wppns[1], ax=axes[1][2], legend='brief', label="Star")  
sns.lineplot(y=sf_erg_wppns[0], x=sf_erg_wppns[1], ax=axes[1][2], legend='brief', label="ERG")  

sns.despine()

plt.savefig("cpr_cppn_wpr_wppn.png", dpi=200)
# plt.show()


#!/usr/bin/env python
"""
    Implementation of the Ricart-Agrawala Algorithm for mutual exclusion.
"""

__author__ = "Berker Acır"
__contact__ = "berkeracir159@gmail.com"
__copyright__ = "Copyright 2021, WINSLAB"
__credits__ = ["Berker Acır"]
__date__ = "2021/04/24"
__deprecated__ = False
__email__ = "berkeracir159@gmail.com"
__license__ = "GPLv3"
__maintainer__ = "developer"
__status__ = "Production"
__version__ = "0.0.1"

from enum import Enum
from time import sleep

from ahc.Ahc import ComponentModel, Event, EventTypes, GenericMessage, GenericMessagePayload, GenericMessageHeader, inf, \
    Topology


class RicartAgrawalaEventTypes(Enum):
    REQUEST = "REQUEST"
    REPLY = "REPLY"
    PRIVILEGE = "PRIVILEGE"


class RicartAgrawalaMessageTypes(Enum):
    REQUEST = "REQUEST"
    REPLY = "REPLY"


class RicartAgrawalaMessageHeader(GenericMessageHeader):

    def __init__(self, messageType, messageFrom, messageTo, nextHop=float('inf'), interfaceID=float('inf'),
                 sequenceID=-1):
        super().__init__(messageType, messageFrom, messageTo, nextHop, interfaceID, sequenceID)


class RicartAgrawalaMessagePayload(GenericMessagePayload):

    def __init__(self, clock, nodeID):
        self.clock = clock
        self.nodeID = nodeID


class MutualExclusionAgrawalaComponent(ComponentModel):
    privilegeSleepAmount = 1

    def __init__(self, componentName, componentID):
        super().__init__(componentName, componentID)

        self.eventhandlers[RicartAgrawalaEventTypes.REQUEST] = self.request_received
        self.eventhandlers[RicartAgrawalaEventTypes.REPLY] = self.reply_received
        self.eventhandlers[RicartAgrawalaEventTypes.PRIVILEGE] = self.on_privilege

        self.clock = 0
        self.havePendingRequest = False
        self.pendingRequestClock = None
        self.isPrivileged = False

        self.deferredRequests = list()
        self.receivedReplies = set()
        self.otherNodeIDs = set()

        self.privilegeCount = 0
        self.sentRequestCount = 0
        self.sentReplyCount = 0
        self.receivedRequestCount = 0
        self.receivedReplyCount = 0
        self.forwardedMessageCount = 0

    def on_init(self, eventobj: Event):
        self.otherNodeIDs = set(Topology().nodes.keys())
        self.otherNodeIDs.remove(self.componentinstancenumber)

    def on_message_from_bottom(self, eventobj: Event):
        message = eventobj.eventcontent
        header = message.header
        messageType = header.messagetype
        messageTo = header.messageto

        if messageTo == self.componentinstancenumber:
            if messageType == RicartAgrawalaMessageTypes.REQUEST:
                eventobj.event = RicartAgrawalaEventTypes.REQUEST
                self.send_self(eventobj)
            elif messageType == RicartAgrawalaMessageTypes.REPLY:
                eventobj.event = RicartAgrawalaEventTypes.REPLY
                self.send_self(eventobj)
        else:
            nextHop = Topology().get_next_hop(self.componentinstancenumber, messageTo)
            interfaceID = f"{self.componentinstancenumber}-{nextHop}"

            if nextHop != inf and nextHop != self.componentinstancenumber:
                self.forwardedMessageCount += 1
                header.nexthop = nextHop
                header.interfaceid = interfaceID
                self.send_down(Event(self, EventTypes.MFRT, message))

    def request_received(self, eventobj: Event):
        self.receivedRequestCount += 1
        receivedRequestClock = eventobj.eventcontent.payload.clock
        receivedRequestNodeID = eventobj.eventcontent.payload.nodeID

        if not self.isPrivileged:   # Not privileged
            if self.havePendingRequest:     # Have pending request
                if receivedRequestClock < self.pendingRequestClock:
                    isMessageDeferred = False
                elif receivedRequestClock == self.pendingRequestClock and receivedRequestNodeID < self.componentinstancenumber:
                    isMessageDeferred = False
                else:
                    isMessageDeferred = True
            else:   # Does not have pending request
                isMessageDeferred = False
        else:   # Privileged
            isMessageDeferred = True

        if isMessageDeferred:
            self.deferredRequests.append(eventobj)
        else:
            if self.clock <= receivedRequestClock:
                self.clock = receivedRequestClock + 1
            self.send_reply(receivedRequestNodeID)

    def reply_received(self, eventobj: Event):
        self.receivedReplyCount += 1
        replyFrom = eventobj.eventcontent.payload.messagepayload
        self.receivedReplies.add(replyFrom)

        if len(self.receivedReplies) == len(self.otherNodeIDs):
            self.send_self(Event(self, RicartAgrawalaEventTypes.PRIVILEGE, None))
        elif len(self.receivedReplies) > len(self.otherNodeIDs):
            raise RuntimeError("Received reply message count exceeded expected limit!")

    def on_privilege(self, eventobj: Event):
        self.isPrivileged = True
        self.havePendingRequest = False
        self.receivedReplies.clear()

        self.privilegeCount += 1
        sleep(self.privilegeSleepAmount)

        self.isPrivileged = False
        self.send_replies_to_deferred_requests()

    def send_request(self):
        self.sentRequestCount += 1
        self.havePendingRequest = True
        self.pendingRequestClock = self.clock
        self.clock += 1

        for nodeID in self.otherNodeIDs:
            nextHop = Topology().get_next_hop(self.componentinstancenumber, nodeID)
            interfaceID = f"{self.componentinstancenumber}-{nextHop}"
            header = RicartAgrawalaMessageHeader(RicartAgrawalaMessageTypes.REQUEST, self.componentinstancenumber,
                                                 nodeID, nextHop, interfaceID)
            payload = RicartAgrawalaMessagePayload(self.pendingRequestClock, self.componentinstancenumber)
            message = GenericMessage(header, payload)
            self.send_down(Event(self, EventTypes.MFRT, message))

    def send_reply(self, nodeID):
        self.sentReplyCount += 1
        nextHop = Topology().get_next_hop(self.componentinstancenumber, nodeID)
        interfaceID = f"{self.componentinstancenumber}-{nextHop}"
        header = RicartAgrawalaMessageHeader(RicartAgrawalaMessageTypes.REPLY, self.componentinstancenumber, nodeID,
                                             nextHop, interfaceID)
        payload = GenericMessagePayload(self.componentinstancenumber)
        message = GenericMessage(header, payload)
        self.send_down(Event(self, EventTypes.MFRT, message))

    def send_replies_to_deferred_requests(self):
        while self.deferredRequests:
            event = self.deferredRequests.pop(0)
            deferredRequestNodeID = event.eventcontent.payload.nodeID
            deferredRequestClock = event.eventcontent.payload.clock

            if self.clock <= deferredRequestClock:
                self.clock = deferredRequestClock + 1
            self.send_reply(deferredRequestNodeID)



import time
import pickle
import random
import pandas as pd
import seaborn as sns
import networkx as nx
import matplotlib.pyplot as plt
from datetime import datetime as dt

from ahc.Ahc import Topology
from graph import ERG, Grid, Star
from ahc.TerminationDetection.shavit_francez import ShavitFrancezAdHocNode, SFAHCNodeSimulationStatus
from ahc.Channels.Channels import P2PFIFOPerfectChannel

def run_shavit_francez_simulation(args):
    if args.network_type == "erg":
        N = ERG(args.node_count, args.node_connectivity)
        total_nodes = args.node_count
    elif args.network_type == "grid":
        N = Grid(args.node_count_on_edge)
        total_nodes = args.node_count_on_edge ** 2
    elif args.network_type == "star":
        N = Star(args.slave_count, master_is_root=args.master_is_root)
        total_nodes = args.slave_count + 1

    if args.run_until_termination:
        args.simulation_ticks = 10**10

    assert args.hard_stop_max_tick < args.simulation_ticks
    assert args.hard_stop_min_tick > 0

    hard_stop_on_tick = None

    if args.hard_stop_nodes:
        hard_stop_on_tick = [random.randint(args.hard_stop_min_tick, args.hard_stop_max_tick) for _ in range(total_nodes)]

    node_active_ticks_initial = [random.randint(args.node_min_activeness_after_receive, args.node_max_activeness_after_receive) if random.random() <= args.node_initial_activeness_prob else 0 for _ in range(total_nodes)]
    alive_nodes = list(range(total_nodes))
    
    topo_context = {
        "network": N,
        "ms_per_tick": args.ms_per_tick,
        "simulation_ticks": args.simulation_ticks,
        "initial_liveness": node_active_ticks_initial,
        "communication_on_active_prob": args.node_activeness_communication_prob,
        "min_activeness_after_receive": args.node_min_activeness_after_receive,
        "max_activeness_after_receive": args.node_max_activeness_after_receive,
        "node_package_process_per_tick": args.node_package_process_per_tick,
        "passiveness_death_thresh": args.passiveness_death_thresh,
        "hard_stop_on_tick": hard_stop_on_tick,
        "alive_nodes": alive_nodes,
        "only_root_alive_initially": args.only_root_alive_initially
    }

    print(topo_context)
    # N.plot()
    # plt.show()

    topo = Topology()
    topo.construct_from_graph(N.G, ShavitFrancezAdHocNode, P2PFIFOPerfectChannel, context=topo_context)

    topo.start()

    stats = {
        "df": pd.DataFrame(data={
            "dead_nodes": [],
            "active_nodes": [],
            "packets_in_transmit": [],
            "queued_packets": [],
            "control_packets_sent": [],
            "control_basic_instant_ratio": [],
            "control_total_cumulative_ratio": [],
            "wave_packets_sent": [],
            "wave_basic_instant_ratio": [],
            "wave_total_cumulative_ratio": [],
            "wave_control_cumulative_ratio": []
        }),
        "terminated_on_tick": None,
        "announced_on_tick": None
    }

    fig, axes = plt.subplots(2, 3)
    fig.set_figwidth(25)
    fig.set_figheight(10)
    # fig.tight_layout()

    input("\n>>> Proceed ?")

    term_wait_ctr = 0
    reason = None
    wave_finisher = None

    try:
        for t in range(1, args.simulation_ticks + 1):
            print(f"[S] Tick: {t}")

            packages_sent = 0
            packages_waiting_on_queue = 0
            num_nodes_active = 0
            num_dead_nodes = 0
            break_this_tick = False
            control_packets_sent = 0
            wave_packets_sent = 0

            node_color = []

            for index, node in topo.nodes.items():
                new_state, pkg_sent_to_friend, cps, wps = node.simulation_tick()

                if N.root == index and new_state != SFAHCNodeSimulationStatus.OUT_OF_TREE:
                    node_color.append("red")
                elif N.root == index and new_state == SFAHCNodeSimulationStatus.OUT_OF_TREE:
                    node_color.append("orange")
                elif new_state == SFAHCNodeSimulationStatus.ACTIVE:
                    node_color.append("green")
                elif new_state == SFAHCNodeSimulationStatus.PASSIVE:
                    node_color.append("mediumslateblue")
                elif new_state == SFAHCNodeSimulationStatus.OUT_OF_TREE:
                    node_color.append("gray")
                else: # is None...
                    # It's the wave finisher!!!
                    node_color.append("yellow")
                    wave_finisher = index

                if new_state is None:
                    reason = f"wave terminated ({t}) ({wave_finisher})"
                    break_this_tick = True

                if new_state == SFAHCNodeSimulationStatus.ACTIVE:
                    num_nodes_active += 1
                elif new_state == SFAHCNodeSimulationStatus.OUT_OF_TREE:
                    num_dead_nodes += 1

                if pkg_sent_to_friend is not None:
                    packages_sent += 1

                control_packets_sent += cps
                wave_packets_sent += wps
                packages_waiting_on_queue += node.waiting_packages_on_queue

            print(f"   (ACTIVE: {num_nodes_active}, PKGS-WAIT: {packages_waiting_on_queue}, PKGS-SENT: {packages_sent})")

            if (packages_waiting_on_queue == 0 and num_nodes_active == 0 and packages_sent == 0):
                if stats["terminated_on_tick"] is None: 
                    stats["terminated_on_tick"] = t

                print("!!! TERMINATED !!!")

                term_wait_ctr += 1

                if args.wait_ticks_after_termination > 0 and term_wait_ctr > args.wait_ticks_after_termination:
                    print("!!! FORCE TERMINATED !!!")
                    reason = f"forced ({args.wait_ticks_after_termination})"
                    break_this_tick = True

                if args.exit_on_termination:
                    break_this_tick = True

            control_pkgs_sent_cum = stats["df"]["control_packets_sent"].sum() + control_packets_sent
            total_pkgs_sent_cum = (stats["df"]["control_packets_sent"].sum() + control_packets_sent + stats["df"]["packets_in_transmit"].sum() + packages_sent + stats["df"]["wave_packets_sent"].sum() + wave_packets_sent)

            stats["df"].loc[t-1] = [
                num_dead_nodes,
                num_nodes_active,
                packages_sent,
                packages_waiting_on_queue,
                control_packets_sent,
                (control_packets_sent / packages_sent) if packages_sent > 0 else 0, # TODO: Fix later, find a better soln,,,
                (((stats["df"]["control_packets_sent"].sum() + control_packets_sent) / total_pkgs_sent_cum) if total_pkgs_sent_cum > 0 else 0) * 100,
                wave_packets_sent,
                (wave_packets_sent / packages_sent) if packages_sent > 0 else 0, # TODO: Fix later, find a better soln,,,
                (((stats["df"]["wave_packets_sent"].sum() + wave_packets_sent) / total_pkgs_sent_cum) if total_pkgs_sent_cum > 0 else 0) * 100,
                (((stats["df"]["wave_packets_sent"].sum() + wave_packets_sent) / control_pkgs_sent_cum) if control_pkgs_sent_cum > 0 else 0) * 100,
            ]

            # axes.scatter(x=t, y=num_nodes_active)
            
            if not args.no_realtime_plot:
                axes[0][0].cla()
                axes[0][1].cla()
                axes[0][2].cla()
                axes[1][0].cla()
                axes[1][1].cla()
                axes[1][2].cla()

                sns.lineplot(data=stats["df"]["active_nodes"], ax=axes[0][0], color="orange")
                sns.lineplot(data=stats["df"]["dead_nodes"], ax=axes[0][0], color="blue")
                sns.lineplot(data=stats["df"]["packets_in_transmit"], ax=axes[0][1], color="purple")
                sns.lineplot(data=stats["df"]["control_packets_sent"], ax=axes[0][1], color="green")
                sns.lineplot(data=stats["df"]["wave_packets_sent"], ax=axes[0][1], color="blue")
                sns.lineplot(data=stats["df"]["control_total_cumulative_ratio"], ax=axes[1][0], color="mediumslateblue")
                sns.lineplot(data=stats["df"]["wave_total_cumulative_ratio"], ax=axes[1][0], color="yellow")
                sns.lineplot(data=stats["df"]["control_basic_instant_ratio"], ax=axes[1][1], color="red")
                sns.lineplot(data=stats["df"]["wave_basic_instant_ratio"], ax=axes[1][1], color="orange")
                sns.lineplot(data=stats["df"]["wave_control_cumulative_ratio"], ax=axes[0][2], color="mediumslateblue")
                
                if args.network_type == "grid":   
                    pos = {i: (i // args.node_count_on_edge, i % args.node_count_on_edge) for i in range(total_nodes)}
                    nx.draw(N.G, ax=axes[1][2], with_labels=True, node_color=node_color, pos=pos)
                else:
                    nx.draw(N.G, ax=axes[1][2], with_labels=True, node_color=node_color)

                plt.pause(0.0005)

            if break_this_tick:
                break

            time.sleep(args.ms_per_tick / 1000)
    except KeyboardInterrupt:
        pass

    ts = dt.now().timestamp()

    if args.no_realtime_plot:
        axes[0][0].cla()
        axes[0][1].cla()
        axes[0][2].cla()
        axes[1][0].cla()
        axes[1][1].cla()
        axes[1][2].cla()

        sns.lineplot(data=stats["df"]["active_nodes"], ax=axes[0][0], color="orange")
        sns.lineplot(data=stats["df"]["dead_nodes"], ax=axes[0][0], color="blue")
        sns.lineplot(data=stats["df"]["packets_in_transmit"], ax=axes[0][1], color="purple")
        sns.lineplot(data=stats["df"]["control_packets_sent"], ax=axes[0][1], color="green")
        sns.lineplot(data=stats["df"]["wave_packets_sent"], ax=axes[0][1], color="blue")
        sns.lineplot(data=stats["df"]["control_total_cumulative_ratio"], ax=axes[1][0], color="mediumslateblue")
        sns.lineplot(data=stats["df"]["wave_total_cumulative_ratio"], ax=axes[1][0], color="yellow")
        sns.lineplot(data=stats["df"]["control_basic_instant_ratio"], ax=axes[1][1], color="red")
        sns.lineplot(data=stats["df"]["wave_basic_instant_ratio"], ax=axes[1][1], color="orange")
        sns.lineplot(data=stats["df"]["wave_control_cumulative_ratio"], ax=axes[0][2], color="mediumslateblue")
        
        if args.network_type == "grid":   
            pos = {i: (i // args.node_count_on_edge, i % args.node_count_on_edge) for i in range(total_nodes)}
            nx.draw(N.G, ax=axes[1][2], with_labels=True, node_color=node_color, pos=pos)
        else:
            nx.draw(N.G, ax=axes[1][2], with_labels=True, node_color=node_color)

        plt.pause(0.0005)

    plt.savefig(f"simdump/SF_stats_{args.network_type}_{total_nodes}_{ts}.png", dpi=200)

    with open(f"simdump/SF_run_{args.network_type}_{total_nodes}_{ts}.pkl", "wb") as fp:
        pickle.dump({
            "args": args,
            "context": topo_context,
            "stats": stats
        }, fp)

    if not args.no_realtime_plot:
        plt.show()

    print(f"\n{reason} [{t - stats['terminated_on_tick'] if stats['terminated_on_tick'] is not None else None}]")



from ahc.Routing.SSBR.HelperFunctions import messageParser, SSBRRouteSearchMessage, SSBRUnicastMessageFPParser, SSBRUnicastMessage
from ahc.Ahc import ComponentModel, Event, EventTypes, ComponentRegistry

class FP(ComponentModel):
    def __init__(self, componentname, componentid):
        super(FP, self).__init__(componentname, componentid)
        self.componentid = componentid
        self.routingTable = {}
        self.mode = 0
        self.nodeCount = 0
        
    def on_init(self, eventobj: Event):
        print(f"{self.componentname} - #{self.componentid} is up.\n")
        #evt = Event(self, EventTypes.MFRP, "Network interface to peers")
        #self.send_peer(evt)

    def on_message_from_top(self, eventobj: Event):
        if(eventobj.eventcontent.header.messagetype == "UNICASTDATA"):
            evt = Event(self, EventTypes.MFRT, SSBRUnicastMessageFPParser(self, eventobj))
            self.send_down(evt)
        else: 
            evt = Event(self, EventTypes.MFRT, messageParser(self, eventobj))
            self.send_down(evt)
    
    def build_routing_table(self, target, mode = 0, node = 0):
        SelfDRP = ComponentRegistry().get_component_by_key("DRP", self.componentid)
        SelfDRP.routingTableFlag = True
        if mode == 1:
            self.mode = 1
            self.nodeCount = node
        evt = Event(self, EventTypes.MFRT, SSBRRouteSearchMessage(self, target))
        self.send_down(evt)
        
    def on_message_from_peer(self, eventobj: Event):
        
        if eventobj.eventcontent.header.messagetype == "ROUTESEARCH":
            evt = Event(self, EventTypes.MFRT,messageParser(self,eventobj))
            self.send_down(evt)

        elif eventobj.eventcontent.header.messagetype == "ROUTEREPLY":
            payload = eventobj.eventcontent.payload
            
            for el in payload:
                    self.routingTable[str(el)] = payload[len(payload)-1]

            if int(eventobj.eventcontent.header.messageto.split("-")[1]) == self.componentid:
                print("Routing table is completed....\n")
                if self.mode == 1:
                    evt = Event(self, EventTypes.MFRB,SSBRUnicastMessage(self, self.nodeCount-1 , "benchmark test message"))
                    self.send_up(evt)

            else:
                evt = Event(self, EventTypes.MFRT,messageParser(self,eventobj))
                self.send_down(evt)
            
        else:
            evt = Event(self, EventTypes.MFRB,messageParser(self,eventobj))
            self.send_up(evt) # send incoming messages to upper components
        
    def printRoutingTable(self):
        print(self.routingTable)

    def getRoutingTable(self):
        return self.routingTable





import sys

from ahc.Ahc import ComponentModel, EventTypes, ConnectorTypes
from ahc.Ahc import Event

from ahc.Consensus.Raft.states import Follower
import logging
logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

logger = logging.getLogger(__name__)


class RaftConsensusComponent(ComponentModel):

    def __init__(self, componentname, componentinstancenumber):
        super().__init__(componentname, componentinstancenumber, 1)
        self.state = None

    def on_message_from_bottom(self, eventobj: Event):
        self.data_received_peer(eventobj.eventsource, eventobj.eventcontent)

    def on_message_from_top(self, eventobj: Event):
        print("client asking for a trouble...")
        self.data_received_client(eventobj.eventsource, eventobj.eventcontent)

    def change_state(self, new_state):
        self.state.teardown()
        logger.info('For component %s State change:' + new_state.__name__, self.componentinstancenumber)
        self.state = new_state(old_state=self.state)

    def data_received_peer(self, sender, message):
        self.state.data_received_peer(sender, message)

    def data_received_client(self, client, message):
        self.state.data_received_client(client, message)

    def send(self, client, message):
        client.send(message)

    def send_to_component(self, recipient, message):
        if recipient != self:
            for conn in self.connectors[ConnectorTypes.DOWN]:
                if conn.componentinstancenumber.find(recipient.componentinstancenumber) != -1:
                    conn.trigger_event(Event(self, EventTypes.MFRT, message))

    def broadcast_peers(self, message):
        self.send_down(Event(self, EventTypes.MFRT, message))

    def on_init(self, eventobj: Event):
        self.state = Follower(server=self)



#!/usr/bin/env python

"""
Implementation of the "IEEE1394 (FireWire)" as described in the textbook
"Fokkink, Wan. Distributed algorithms: an intuitive approach. MIT Press,
2018.", first introduced in " IEEE 1394-1995 - IEEE Standard for a High
Performance Serial Bus"
"""

__author__ = "Yigit Sever"
__contact__ = "yigit@yigitsever.com"
__copyright__ = "Copyright 2021, WINSLAB"
__credits__ = ["Yigit Sever"]
__date__ = "2021-05-24"
__deprecated__ = False
__email__ = "yigit@yigitsever.com"
__license__ = "GPLv3"
__maintainer__ = "developer"
__status__ = "Production"
__version__ = "0.0.1"

import datetime
import random
from enum import Enum
from time import sleep

from ahc.Ahc import (ComponentModel, Event, EventTypes, GenericMessage,
                 GenericMessageHeader, GenericMessagePayload, Topology)


class FireWirePacketType(Enum):
    """Two types of FireWire requests:
    - Parent Request
    - Acknowledgement"""

    PARENT_REQ = "PARENT_REQ"
    ACKNOWLEDGEMENT = "ACKNOWLEDGEMENT"
    START_TIMER = "START_TIMER"
    CHECK_TIMER = "CHECK_TIMER"
    TIMEOUT = "TIMEOUT"
    ROOT_CONTENTION = "ROOT_CONTENTION"


class FireWireMessageHeader(GenericMessageHeader):
    def __init__(
        self,
        messagefrom,
        messageto,
        messagetype="FireWire Message",
        nexthop=float("inf"),
        interfaceid=float("inf"),
        sequencenumber=-1,
    ):
        super().__init__(
            messagetype, messagefrom, messageto, nexthop, interfaceid, sequencenumber
        )


class FireWireMessagePayload(GenericMessagePayload):
    def __init__(self):
        super().__init__(messagepayload="FireWire Message")


class FireWireNode(ComponentModel):

    # For animation/plotting
    callback = None
    draw_delay = None

    def __init__(self, component_name, component_id):
        super().__init__(component_name, component_id)
        self.eventhandlers[FireWirePacketType.START_TIMER] = self.on_timer_initialize
        self.eventhandlers[FireWirePacketType.CHECK_TIMER] = self.check_timer
        self.eventhandlers[FireWirePacketType.TIMEOUT] = self.timeout
        self.eventhandlers[FireWirePacketType.ROOT_CONTENTION] = self.root_contention

        self.parent = None
        self.received = list()
        self.neighbours = set()
        self.is_leader = False
        self.in_root_contention = False
        self.is_waiting = False
        self.is_terminated = False

        self.waiting_since = None
        self.timeout_duration = 2

    def on_init(self, eventobj: Event):
        sleep(1)
        self.neighbours = set(Topology().get_neighbors(self.componentinstancenumber))
        print(f"the neighbours of {self.componentinstancenumber} is {self.neighbours}")

        self.send_parent_req()

    def send_parent_req(self):
        """Send a parent request to the only eligible neighbour node. The
        neighbour node should not have sent this node a parent request and the
        number of such neighbours of this node should be 1.
        """
        # if there is *only* one possible parent then send them a parent request here

        result = self.neighbours - set(self.received)

        if len(result) == 1:
            par = result.pop()
            self.parent = par

            print(
                f"🤖 {self.componentinstancenumber} picked {self.parent} as it's parent"
            )

            next_hop_interface_id = f"{self.componentinstancenumber}-{self.parent}"

            header = FireWireMessageHeader(
                messagefrom=self.componentinstancenumber,
                messageto=self.parent,
                nexthop=self.parent,
                messagetype=FireWirePacketType.PARENT_REQ,
                interfaceid=next_hop_interface_id
            )
            payload = FireWireMessagePayload()

            message = GenericMessage(header, payload)
            self.send_down(Event(self, EventTypes.MFRT, message))
        else:
            # Cannot send a parent request, more than one possible parent
            return

    def root_contention(self, eventobj: Event):
        if self.is_leader:
            return
        print(f"🤖 {self.componentinstancenumber} is in ROOT CONTENTION")
        decision = random.choice([True, False])

        if decision:
            print(f"🤖 {self.componentinstancenumber} decides to YIELD")
            self.in_root_contention = True
            self.send_parent_req()
            self.is_waiting = False
            self.waiting_since = None
        else:
            print(f"🤖 {self.componentinstancenumber} decides to HOLD")
            self.is_waiting = True
            self.in_root_contention = False
            self.send_self(Event(self, FireWirePacketType.START_TIMER, "..."))

        # self.callback.set()
        # self.draw_delay.wait()
        # self.draw_delay.clear()

    def on_timer_initialize(self, eventobj: Event):
        start_time = eventobj.time
        self.waiting_since = start_time
        self.send_self(Event(self, FireWirePacketType.CHECK_TIMER, "..."))

    def check_timer(self, eventobj: Event):
        current_time = datetime.datetime.now()
        delta = current_time - self.waiting_since
        if delta.seconds > self.timeout_duration:
            self.send_self(Event(self, FireWirePacketType.TIMEOUT, "..."))
        else:
            sleep(0.2)
            self.send_self(Event(self, FireWirePacketType.CHECK_TIMER, "..."))

    def timeout(self, eventobj: Event):
        self.send_self(Event(self, FireWirePacketType.ROOT_CONTENTION, "..."))

    def on_message_from_bottom(self, eventobj: Event):
        """ New message from the link layer """
        header: FireWireMessageHeader = eventobj.eventcontent.header
        # paylaod is not important for FireWire

        if header.messagetype == FireWirePacketType.PARENT_REQ:
            if header.messagefrom is not self.parent:
                new_child = header.messagefrom

                self.received.append(new_child)

                next_hop_interface_id = f"{self.componentinstancenumber}-{new_child}"

                header = FireWireMessageHeader(
                    messagefrom=self.componentinstancenumber,
                    messageto=new_child,
                    nexthop=new_child,
                    messagetype=FireWirePacketType.ACKNOWLEDGEMENT,
                    interfaceid=next_hop_interface_id
                )
                payload = FireWireMessagePayload()

                ack = GenericMessage(header, payload)
                self.send_down(Event(self, EventTypes.MFRT, ack))

                self.send_parent_req()
            elif not self.is_waiting:
                self.send_self(Event(self, FireWirePacketType.ROOT_CONTENTION, "..."))
            else:
                print(f" 👑 {self.componentinstancenumber} is elected as the leader")
                self.is_leader = True
                self.is_terminated = True

        elif (
            header.messagetype == FireWirePacketType.ACKNOWLEDGEMENT
            and header.messagefrom == self.parent
        ):
            # This node's parent request got acknowledged, the process can
            # safely terminate
            print(
                f"🤖 {self.componentinstancenumber} received an ACK "
                f" from {header.messagefrom}, terminating"
            )
            self.is_terminated = True
            self.in_root_contention = False

        self.callback.set()
        self.draw_delay.wait()
        self.draw_delay.clear()



from heapq import heapify, heappush, heappop
from typing import Dict


class MinHeapNode:
    def __init__(self, key: int, data: Dict):
        self.key = key
        self.data = data

    def __eq__(self, o: object) -> bool:
        return self.key == o.key

    def __ne__(self, o: object) -> bool:
        return not self.key == o.key

    def __le__(self, other):
        return self == other or self < other

    def __lt__(self, other):
        return self.data['d'] < other.data['d']

    def __ge__(self, other):
        return self == other or self > other

    def __gt__(self, other):
        return self.data['d'] > other.data['d']

    def __str__(self):
        return f"#{self.key} / {self.data}"

    def __repr__(self):
        return f"#{self.key} / {self.data}"


class MinHeap:
    def __init__(self):
        self.heap = []
        heapify(self.heap)

    def insert(self, node):
        if node not in self.heap:
            heappush(self.heap, node)

    def print(self):
        print("The heap elements: ")

        for idx, i in enumerate(self.heap):
            print(f"{idx} => {i}")

    def extract_min(self) -> MinHeapNode:
        return heappop(self.heap)

    def is_empty(self) -> bool:
        return len(self.heap) == 0



from ahc.Ahc import ComponentModel
from ahc.Ahc import ComponentRegistry
from ahc.Ahc import EventTypes
from ahc.Ahc import Event
from ahc.Ahc import GenericMessage
from ahc.Ahc import GenericMessageHeader



class ApplicationComponent(ComponentModel):
    def __init__(self, component_name, component_id):
        super(ApplicationComponent, self).__init__(component_name, component_id)

    def find_total_node_number(self) -> int:
        count = 0
        try:
            while True:
                ComponentRegistry().get_component_by_key(self.componentname, count)
                count = count + 1
        except KeyError:
            pass

        return count

    def send_data(self, dst):

        src = self.componentinstancenumber

        data = "Data send by " + \
               self.componentname + \
               "-" +\
               str(self.componentinstancenumber)

        message_header = GenericMessageHeader("", self.componentname + "-" + str(src),
                                              self.componentname + "-" + str(dst))
        message_payload = data
        message = GenericMessage(message_header, message_payload)

        event = Event(self, EventTypes.MFRT, message)
        self.send_down(event)

    def on_init(self, eventobj: Event):
        if self.componentinstancenumber == 0:
            last_components_id = self.find_total_node_number() - 1
            dst = last_components_id
            self.send_data(dst)

    def process_message(self, data: str) -> None:
        print(f"I am {self.componentname}-{self.componentinstancenumber}, data received = {data}\n")

    def on_message_from_bottom(self, eventobj: Event):
        data = eventobj.eventcontent.payload
        self.process_message(data)



from ahc.Ahc import Lock


def singleton(cls):
  instance = [None]
  def wrapper(*args, **kwargs):
    if instance[0] is None:
      instance[0] = cls(*args, **kwargs)
    return instance[0]
  return wrapper

@singleton
class GSRExperimentCollector:
    MESSAGE_COUNT = {}
    completion = []
    first_routing_completion = 0
    n_control_messages = 0
    completion_lock = Lock()



import enum
import threading
import time
from copy import deepcopy

from ahc.Ahc import ComponentModel
from ahc.Ahc import Event
from ahc.Ahc import EventTypes
from ahc.Ahc import GenericMessage
from ahc.Ahc import GenericMessageHeader
from ahc.Ahc import Thread


class MessageTypes(enum.Enum):
    ROUTE_FORWARDING = "ROUTE_FORWARDING"
    ROUTE_ERROR = "ROUTE_ERROR"
    ROUTE_REPLY = "ROUTE_REPLY"
    ROUTE_REQUEST = "ROUTE_REQUEST"


class Cache:

    def __init__(self, id):
        self.id = id
        self.lock = threading.RLock()
        self.cache = {}

    def has(self, key) -> bool:
        with self.lock:
            if key in self.cache:
                return True

            return False

    def get_value(self, key):
        with self.lock:

            if not self.has(key):
                return None
            else:
                value = self.cache[key]
                return value

    def set_value(self, key, value):
        with self.lock:
            self.cache[key] = value

    def delete_key(self, key):
        with self.lock:
            self.cache.pop(key, None)

    def delete_keys_with_link(self, link):
        with self.lock:

            for key in list(self.cache.keys()):
                value = self.cache[key]
                if value.count(link[0]) > 0 and value.count(link[1]) > 0:
                    if value.index(link[0]) + 1 == value.index(link[1]):
                        self.delete_key(key)


class DSRAlgorithmComponent(ComponentModel):

    def __init__(self, component_name, componentinstancenumber):
        super(DSRAlgorithmComponent, self).__init__(component_name, componentinstancenumber)
        self.hop_limit = 9
        self.uid = 0
        self.trial_number = 2
        self.route_cache = Cache(componentinstancenumber)
        self.route_request_table = Cache(componentinstancenumber)

    def is_destination(self, dst: int) -> bool:
        return dst == self.get_component_id()

    def is_route_request_seen_before(self, src: int, uid: int) -> bool:

        # if request seen before => return true
        # else set route_request_table with new UID

        if self.route_request_table.has(src):
            latest_uid = self.route_request_table.get_value(src)
            if latest_uid is not None and latest_uid >= uid:
                return True

        self.route_request_table.set_value(src, uid)
        return False

    def is_source(self, src: int) -> bool:
        return src == self.get_component_id()

    def get_component_id(self) -> int:
        return self.componentinstancenumber

    @staticmethod
    def get_current_time_in_ms():
        return round(time.time() * 1000)

    def create_app_event(self, src: int, dst: int, data) -> Event:

        # Creates an event to send upper layer with given args

        message_header = GenericMessageHeader("", self.componentname + "-" + str(src),
                                              self.componentname + "-" + str(dst))
        message_payload = data
        message = GenericMessage(message_header, message_payload)

        return Event(self, EventTypes.MFRB, message)

    def create_route_event(self, message_type, src: int, dst: int, route: list, data=None) -> Event:

        # Creates an event to send below layer with given args

        message_header = GenericMessageHeader(message_type, self.componentname + "-" + str(src),
                                              self.componentname + "-" + str(dst))

        new_route = deepcopy(route)
        message_payload = [new_route, data]
        message = GenericMessage(message_header, message_payload)

        return Event(self, EventTypes.MFRT, message)

    def create_unique_id_for_req(self) -> int:
        return self.uid + 1

    def start_data_sending(self, dst: int, data) -> None:

        # send data if dest. route is cached
        # if not try discovery by _trial_number_ times

        for i in range(self.trial_number):

            if self.route_cache.has(dst):
                src = self.get_component_id()
                route = deepcopy(self.route_cache.get_value(dst))

                self.transmit_route_forwarding(src, dst, route, data)
                return

            # waits in start_route_discovery
            self.start_route_discovery(dst)

    def start_route_discovery(self, dst: int) -> None:

        # starts route request phase
        # and wait for reply

        src = self.get_component_id()
        route = [src]
        uid = self.create_unique_id_for_req()

        self.transmit_route_request(src, dst, route, uid)
        self.wait_for_route_reply(dst)

    def start_route_maintenance(self, broken_link: list) -> None:
        self.route_cache.delete_keys_with_link(deepcopy(broken_link))

    def wait_for_route_reply(self, dst: int) -> None:

        # sleeps in a for loop till cache is acquired
        # or timeout occurred

        sleep_period_in_ms = 10  # min for windows
        sleep_period_in_sec = sleep_period_in_ms / 1000

        timeout_in_sec = 100
        timeout_in_ms = timeout_in_sec * 1000
        start_time_in_ms = self.get_current_time_in_ms()
        end_time_in_ms = start_time_in_ms + timeout_in_ms

        while end_time_in_ms > self.get_current_time_in_ms():
            if self.route_cache.has(dst):
                return

            time.sleep(sleep_period_in_sec)

    def add_to_cache(self, src, route):

        # checks cache
        # if dst not exist
        #   add dst route to table
        # else if new route shorter
        #   add dst route to table

        local_route = deepcopy(route)
        if not self.route_cache.has(src):
            self.route_cache.set_value(src, local_route)
        elif len(local_route) < len(self.route_cache.get_value(src)):
            self.route_cache.set_value(src, local_route)

    def on_message_from_top(self, eventobj: Event):

        # creates a thread and starts data sending

        dst = int(eventobj.eventcontent.header.messageto.split("-")[1])
        data = eventobj.eventcontent.payload

        thread = Thread(target=self.start_data_sending, args=[dst, data])
        thread.start()

    def on_message_from_bottom(self, eventobj: Event):

        # filters messages sent below by their message type
        # dispatch them accordingly

        src = int(eventobj.eventcontent.header.messagefrom.split("-")[1])
        dst = int(eventobj.eventcontent.header.messageto.split("-")[1])
        route = deepcopy(eventobj.eventcontent.payload[0])
        data = eventobj.eventcontent.payload[1]

        uid = broken_link = data

        message_type = eventobj.eventcontent.header.messagetype
        if MessageTypes.ROUTE_FORWARDING == message_type:
            self.receive_route_forwarding(src, dst, route, data)

        elif MessageTypes.ROUTE_ERROR == message_type:
            self.receive_route_error(src, dst, route, broken_link)

        elif MessageTypes.ROUTE_REPLY == message_type:
            self.receive_route_reply(src, dst, route)

        elif MessageTypes.ROUTE_REQUEST == message_type:
            self.receive_route_request(src, dst, route, uid)

    def receive_route_forwarding(self, src: int, dst: int, route: list, data) -> None:

        # if destination
        #   send data to above layer
        # else
        #   forward data to next hop

        if self.is_destination(dst):
            event = self.create_app_event(src, dst, data)
            self.send_up(event)
        else:
            self.transmit_route_forwarding(src, dst, deepcopy(route), data)

    def receive_route_error(self, src: int, dst: int, route: list, broken_link: list) -> None:

        # if destination
        #   start route maintenance
        # else
        #   send route error to next node

        if self.is_destination(dst):
            self.start_route_maintenance(deepcopy(broken_link))
        else:
            self.transmit_route_error(src, dst, deepcopy(route), deepcopy(broken_link))

    def receive_route_reply(self, src: int, dst: int, route: list) -> None:

        # if destination
        #   add to cache
        # else
        #   transmit reply packet to next node

        local_route = deepcopy(route)

        if not self.is_destination(dst):
            self.transmit_route_reply(src, dst, local_route)
        else:

            try:
                index_of_current_component = local_route.index(self.get_component_id())

                self.add_to_cache(src, local_route[index_of_current_component:])

            except ValueError:
                print("[DSRAlgorithmComponent:receive_route_reply][Exception] ValueError")
                print(
                    "[DSRAlgorithmComponent:receive_route_reply][Exception] comp_id = " + str(self.get_component_id()))
                str_route = ' '.join(map(str, local_route))
                print("[DSRAlgorithmComponent:receive_route_reply][Exception] route = " + str_route)
                return None

    def receive_route_request(self, src: int, dst: int, route: list, uid: int) -> None:

        # if route request seen before
        #   drop packet
        # else if source node
        #   drop packet
        # else if get_component_id is in route
        #   drop packet
        # else if destination node
        #   add route to cache
        #   start route reply
        # else
        #   if hop limit not exceeded
        #       broadcast route request

        if self.is_route_request_seen_before(src, uid):
            return
        elif self.is_source(src):
            return
        elif self.get_component_id() in route:
            return

        new_route = deepcopy(route)
        new_route.append(self.get_component_id())

        if self.is_destination(dst):

            self.add_to_cache(src, new_route[::-1])
            self.transmit_route_reply(dst, src, new_route)

        else:
            if len(new_route) < self.hop_limit:
                self.transmit_route_request(src, dst, new_route, uid)

    def transmit_route_forwarding(self, src: int, dst: int, route: list, data) -> None:
        event = self.create_route_event(MessageTypes.ROUTE_FORWARDING, src, dst, deepcopy(route), data)
        self.send_down(event)

    def transmit_route_error(self, src: int, dst: int, route: list, broken_link: list) -> None:
        event = self.create_route_event(MessageTypes.ROUTE_ERROR, src, dst, deepcopy(route), broken_link)
        self.send_down(event)

    def transmit_route_reply(self, src: int, dst: int, route: list) -> None:
        event = self.create_route_event(MessageTypes.ROUTE_REPLY, src, dst, deepcopy(route))
        self.send_down(event)

    def transmit_route_request(self, src: int, dst: int, route: list, uid: int) -> None:
        event = self.create_route_event(MessageTypes.ROUTE_REQUEST, src, dst, deepcopy(route), uid)
        self.send_down(event)



#!/usr/bin/env python

"""
Implementation of the "Itai-Rodeh Election Algorithm for Rings" as
described in the textbook "Fokkink, Wan. Distributed algorithms: an intuitive
approach. MIT Press, 2018." with additional help from the paper "W. Fokkink and
J. Pang, Simplifying Itai-Rodeh Leader Election for Anonymous Rings. 2004."
"""

__author__ = "Yigit Sever"
__contact__ = "yigit@yigitsever.com"
__copyright__ = "Copyright 2021, WINSLAB"
__credits__ = ["Yigit Sever"]
__date__ = "2021-05-24"
__deprecated__ = False
__email__ = "yigit@yigitsever.com"
__license__ = "GPLv3"
__maintainer__ = "developer"
__status__ = "Production"
__version__ = "0.0.1"

from enum import Enum
from random import randint

from ahc.Ahc import (ComponentModel, Event, EventTypes, GenericMessage,
                 GenericMessageHeader, GenericMessagePayload, Topology)


class State(Enum):
    """
    State of the nodes, one from {active, passive, leader}
    - "active" nodes are initiators, attempting to become a leader
    - "passive" nodes have selected smaller id's than "active" nodes and cannot
      compete for leadership anymore
    - "leader" node has won an election round, only one such node should be
      present in the network
    """

    active = 1
    passive = 2
    leader = 3


class ItaiRodehMessageHeader(GenericMessageHeader):
    def __init__(
        self,
        messagefrom,
        messageto,
        messagetype="Itai-Rodeh Message",
        nexthop=float("inf"),
        interfaceid=float("inf"),
        sequencenumber=-1,
    ):
        super().__init__(
            messagetype, messagefrom, messageto, nexthop, interfaceid, sequencenumber
        )


class ItaiRodehMessagePayload(GenericMessagePayload):
    """
    Itai-Rodeh Algorithm uses messages with 4 fields (using textbook's
    terminology here):

    - id_p (i): The (random) id the process has chosen for itself for the
      current round
    - election_round (n'): The current election round, old messages are
      silently dismissed
    - hop_count (h): Used so that the originating node can recognize their own
      message
    - dirty_bit (b): For resolving ties, nodes that picked this (highest) id
      will still be active next round
    """

    def __init__(self, election_round, id_p, messagepayload):
        super().__init__(messagepayload)
        self.election_round = election_round
        self.id_p = id_p
        self.hop_count = 1
        # disparity between the paper & textbook, following textbook here
        # mnemonic = bit is not dirty, we can still be the leader
        self.dirty_bit = False


class ItaiRodehNode(ComponentModel):
    """
    Node in a system that uses Itai-Rodeh algorithm
    Each process has three parameters:
    - id_p: 1 <= i <= N where N is the ring size
    - state: from the State enum, active nodes participate in the election,
      passive nodes pass messages around, leader is selected at the end of the
      election cycle
    - round: current election round, starts at 1
    """

    ring_size = 0
    callback = None
    draw_delay = None
    global_round = 1

    def __init__(self, component_name, component_id):
        super().__init__(component_name, component_id)

        """ The anonymous id the node will select for the round """
        self.id_p = 0

        """ Initially, all processes are active """
        self.state = State.active

        """ Initialized at round 0 """
        self.election_round = 1

    def send_election_packet(self):

        header = ItaiRodehMessageHeader(
            messagefrom=self.componentinstancenumber,
            messageto=self.next_hop,
            nexthop=self.next_hop,
            messagetype="ItaiRodeh Message",
            interfaceid=self.next_hop_interface_id
        )
        payload = ItaiRodehMessagePayload(self.election_round, self.id_p)

        message = GenericMessage(header, payload)
        self.send_down(Event(self, EventTypes.MFRT, message))

    def pass_packet_along(self, message):
        """For passive processes

        :message: the whole Event, with eventcontent which includes header and
        payload

        """
        pass

    def on_init(self, eventobj: Event):
        # Select an id for round 1
        self.id_p = randint(1, self.ring_size)
        print(
            f"🤖 {self.componentinstancenumber} selected {self.id_p} as their"
            f" ID for round {self.election_round}"
        )

        # Calculate the neighbour, we're on a directed ring
        self.neighbour_id = (int(self.componentinstancenumber) + 1) % self.ring_size

        self.next_hop = Topology().get_next_hop(
            self.componentinstancenumber, self.neighbour_id
        )

        self.next_hop_interface_id = f"{self.componentinstancenumber}-{self.next_hop}"

        self.send_election_packet()

    def on_message_from_bottom(self, eventobj: Event):
        """ New message from the link layer """
        payload: ItaiRodehMessagePayload = eventobj.eventcontent.payload
        header: ItaiRodehMessageHeader = eventobj.eventcontent.header

        message_election_round = payload.election_round
        message_assumed_id = payload.id_p

        # For the active node, we are going to follow the if/else chain given
        # in the textbook

        if self.state == State.passive:
            # passive node, pass the message on to the next hop, increase the
            # 'hop_count' of packet by one, no other responsibility
            payload.hop_count += 1
            header.messageto = self.next_hop
            header.next_hop = self.next_hop
            header.interfaceid = self.next_hop_interface_id

            message = GenericMessage(header, payload)
            self.send_down(Event(self, EventTypes.MFRT, message))
        elif self.state == State.active:

            if message_election_round > self.election_round or (
                message_election_round == self.election_round
                and message_assumed_id > self.id_p
            ):
                # Another node has picked a higher id than this node for the
                # current round or this node has received a message from a
                # future round, going passive

                print(
                    f"🤖 {self.componentinstancenumber} is PASSIVE: "
                    f"{message_assumed_id} for round {message_election_round} "
                    f"encountered, this node is at {self.election_round} with "
                    f"{self.id_p}"
                )

                self.state = State.passive
                # we clear the id_p here to have them *not* show up in the
                # animation
                self.id_p = " "
                payload.hop_count += 1

                header.messageto = self.next_hop
                header.next_hop = self.next_hop
                header.interfaceid = self.next_hop_interface_id

                message = GenericMessage(header, payload)
                self.send_down(Event(self, EventTypes.MFRT, message))

            elif message_election_round < self.election_round or (
                message_election_round == self.election_round
                and message_assumed_id < self.id_p
            ):
                # This node has received a message from a previous round or
                # from the current round but with a lower assumed id, so this
                # node can dismiss the election attempt of the sender node
                print(
                    f"🤖 {self.componentinstancenumber} is dismissing "
                    f"{message_assumed_id} for round {message_election_round} "
                    f"this node is at {self.election_round} with {self.id_p}"
                )

            elif (
                message_election_round == self.election_round
                and message_assumed_id == self.id_p
            ):

                if payload.hop_count < self.ring_size:
                    # receiver node is not the initial sender node
                    # another node has picked our id, dirty their bit and pass it along
                    payload.dirty_bit = True
                    payload.hop_count += 1

                    header.messageto = self.next_hop
                    header.next_hop = self.next_hop
                    header.interfaceid = self.next_hop_interface_id

                    print(
                        f"🤖 {self.componentinstancenumber} dirtied the bit "
                        f"for {message_assumed_id}, passing it along"
                    )
                    message = GenericMessage(header, payload)
                    self.send_down(Event(self, EventTypes.MFRT, message))

                elif payload.hop_count == self.ring_size:
                    # the message that this node has sent traversed all the way
                    # around the ring
                    print(
                        f"🤖 {self.componentinstancenumber}'s message "
                        f"traversed all the way around"
                    )
                    if payload.dirty_bit:
                        # Bit has been dirtied, next round
                        self.id_p = randint(1, self.ring_size)
                        self.election_round += 1
                        ItaiRodehNode.global_round = self.election_round
                        print(
                            f"🤖 {self.componentinstancenumber} is moving "
                            f"onto round {self.election_round}"
                        )
                        print(
                            f"🤖 {self.componentinstancenumber} selected "
                            f"{self.id_p} as their ID for round "
                            f"{self.election_round}"
                        )
                        self.send_election_packet()
                    else:
                        # The bit is still false, this node is the leader
                        self.state = State.leader
                        print(
                            f"🤖 {self.componentinstancenumber}: I'M THE ELECTED LEADER"
                        )
        self.callback.set()
        self.draw_delay.wait()
        self.draw_delay.clear()



import time
from enum import Enum

from ahc.Ahc import ComponentModel
from ahc.Ahc import Event
from ahc.Ahc import GenericMessagePayload, GenericMessageHeader, GenericMessage, EventTypes
from ahc.Ahc import MessageDestinationIdentifiers

#################FAILURE ASSUMPTIONS
# TODO: Correctness properties: Safety and liveness
# TODO: Model fail stop, fail silent, omission, fail-recover and byzantine failure models to the components
# TODO: Add storage (temporary, persistent) to crash-recover model

# define your own message types
class FailureDetectorMessageTypes(Enum):
  IAMALIVE = "IAMALIVE"

# define your own message header structure
class FailureDetectorMessageHeader(GenericMessageHeader):
  pass

# define your own message payload structure
class FailureDetectorMessagePayload(GenericMessagePayload):
  pass

# define your own component model extending the generic component model
# do not forget to define onInit event handler...MUST!

class FailureDetector(ComponentModel):
  def on_tx_alive_message(self, eventobj: Event):
    time.sleep(self.alivemessageperiod)  # Period of alive messages
    # Send down the I'm Alive mesage
    # print("I am alive....")
    hdr = FailureDetectorMessageHeader(FailureDetectorMessageTypes.IAMALIVE, self.componentinstancenumber,
                                       MessageDestinationIdentifiers.LINKLAYERBROADCAST)
    payload = FailureDetectorMessagePayload(f"I am Node.{self.componentinstancenumber} and I am live ")
    failuredetectormessage = GenericMessage(hdr, payload)
    self.send_down(Event(self, EventTypes.MFRT, failuredetectormessage))
    # Schedule the next I'm Alive message
    self.send_self(Event(self, "txalivemessage", "timer for alive message"))

  def on_message_from_bottom(self, eventobj: Event):
    try:
      failuredetectormessage = eventobj.eventcontent
      hdr = failuredetectormessage.header
      if hdr.messagetype == FailureDetectorMessageTypes.IAMALIVE:
        print(f"Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} has sent {hdr.messagetype} message")
      else:
        print(f"Node-{self.componentinstancenumber} says received {hdr.messagetype}")

    except AttributeError:
      print("Attribute Error")

  def on_init(self, eventobj: Event):
    self.alivemessageperiod = 1
    self.send_self(Event(self, "txalivemessage", "timer for alive message"))

  def __init__(self, componentname, componentinstancenumber):
    super().__init__(componentname, componentinstancenumber)
    self.eventhandlers["txalivemessage"] = self.on_tx_alive_message



#!/usr/bin/env python
"""
    Implementation of the Raymond's Algorithm for mutual exclusion.
"""

__author__ = "Berker Acır"
__contact__ = "berkeracir159@gmail.com"
__copyright__ = "Copyright 2021, WINSLAB"
__credits__ = ["Berker Acır"]
__date__ = "2021/05/18"
__deprecated__ = False
__email__ = "berkeracir159@gmail.com"
__license__ = "GPLv3"
__maintainer__ = "developer"
__status__ = "Production"
__version__ = "0.0.1"

from enum import Enum
from time import sleep
import networkx as nx

from ahc.Ahc import ComponentModel, Event, EventTypes, GenericMessage, GenericMessagePayload, GenericMessageHeader, Topology


class RaymondEventTypes(Enum):
    TOKEN = "TOKEN"
    REQUEST = "REQUEST"
    PRIVILEGE = "PRIVILEGE"


class RaymondMessageTypes(Enum):
    TOKEN = "TOKEN"
    REQUEST = "REQUEST"


class RaymondMessageHeader(GenericMessageHeader):

    def __init__(self, messageType, messageFrom, messageTo, nextHop=float('inf'), interfaceID=float('inf'),
                 sequenceID=-1):
        super().__init__(messageType, messageFrom, messageTo, nextHop, interfaceID, sequenceID)


class RaymondMessagePayload(GenericMessagePayload):

    def __init__(self, nodeID):
        self.nodeID = nodeID


class MutualExclusionRaymondComponent(ComponentModel):
    privilegeSleepAmount = 1

    def __init__(self, componentName, componentID):
        super().__init__(componentName, componentID)

        self.eventhandlers[RaymondEventTypes.TOKEN] = self.token_received
        self.eventhandlers[RaymondEventTypes.REQUEST] = self.request_received
        self.eventhandlers[RaymondEventTypes.PRIVILEGE] = self.on_privilege

        self.neighborNodeIDs = set()
        self.parentNodeID = None
        self.queue = list()

        self.isRoot = False
        self.havePendingRequest = False
        self.isPrivileged = False

        self.privilegeCount = 0
        self.sentRequestCount = 0
        self.sentTokenCount = 0
        self.receivedRequestCount = 0
        self.receivedTokenCount = 0
        self.forwardedMessageCount = 0

    def on_init(self, eventobj: Event):
        mstG = nx.minimum_spanning_tree(Topology().G)
        self.neighborNodeIDs = set(mstG.neighbors(self.componentinstancenumber))
        if self.componentinstancenumber == 0:
            self.isRoot = True
        else:
            self.parentNodeID = nx.shortest_path(mstG, self.componentinstancenumber, 0)[1]

    def on_message_from_bottom(self, eventobj: Event):
        message = eventobj.eventcontent
        header = message.header
        messageType = header.messagetype
        messageTo = header.messageto

        if messageTo == self.componentinstancenumber:
            if messageType == RaymondMessageTypes.REQUEST:
                eventobj.event = RaymondEventTypes.REQUEST
                self.send_self(eventobj)
            elif messageType == RaymondMessageTypes.TOKEN:
                eventobj.event = RaymondEventTypes.TOKEN
                self.send_self(eventobj)

    def put(self, nodeID=None):
        if nodeID is None:
            nodeID = self.componentinstancenumber

        if not self.queue:
            headChanged = True
        else:
            headChanged = False
        self.queue.append(nodeID)
        if nodeID == self.componentinstancenumber:
            self.havePendingRequest = True

        if headChanged:
            if self.isRoot:
                if nodeID == self.componentinstancenumber:
                    self.send_self(Event(self, RaymondEventTypes.PRIVILEGE, None))
                else:
                    self.send_token(nodeID)
            else:
                self.send_request()

    def pop(self):
        nodeID = self.queue.pop(0)
        if nodeID == self.componentinstancenumber and nodeID not in self.queue:
            self.havePendingRequest = False

        if self.queue:
            head = self.queue[0]
            if self.isRoot:
                if head == self.componentinstancenumber:
                    self.send_self(Event(self, RaymondEventTypes.PRIVILEGE, None))
                else:
                    self.send_token(head)
            else:
                self.send_request()

    def on_privilege(self, eventobj: Event):
        self.isPrivileged = True

        self.privilegeCount += 1
        sleep(self.privilegeSleepAmount)

        self.isPrivileged = False
        self.pop()

    def token_received(self, eventobj: Event):
        self.receivedTokenCount += 1
        self.isRoot = True
        self.parentNodeID = None

        head = self.queue[0]
        if head == self.componentinstancenumber:
            self.send_self(Event(self, RaymondEventTypes.PRIVILEGE, None))
        else:
            self.send_token(head)

    def request_received(self, eventobj: Event):
        self.receivedRequestCount += 1
        receivedRequestNodeID = eventobj.eventcontent.payload.nodeID
        self.put(receivedRequestNodeID)

    def send_token(self, nodeID):
        self.sentTokenCount += 1
        self.isRoot = False
        self.parentNodeID = nodeID

        nextHop = self.parentNodeID
        interfaceID = f"{self.componentinstancenumber}-{nextHop}"
        header = RaymondMessageHeader(RaymondMessageTypes.TOKEN, self.componentinstancenumber, self.parentNodeID,
                                      nextHop, interfaceID)
        payload = RaymondMessagePayload(self.componentinstancenumber)
        message = GenericMessage(header, payload)
        self.send_down(Event(self, EventTypes.MFRT, message))

        self.pop()

    def send_request(self):
        self.sentRequestCount += 1
        nextHop = self.parentNodeID
        interfaceID = f"{self.componentinstancenumber}-{nextHop}"
        header = RaymondMessageHeader(RaymondMessageTypes.REQUEST, self.componentinstancenumber, self.parentNodeID,
                                      nextHop, interfaceID)
        payload = RaymondMessagePayload(self.componentinstancenumber)
        message = GenericMessage(header, payload)
        self.send_down(Event(self, EventTypes.MFRT, message))



from ahc.Ahc import ComponentModel, Event, GenericMessage, GenericMessageHeader, EventTypes

class DataSender(ComponentModel):
    def __init__(self, componentname, componentid):
        super(DataSender, self).__init__(componentname, componentid)
        pass

    def on_init(self, eventobj: Event):
        pass

    def on_message_from_bottom(self, eventobj: Event):
        message_target = eventobj.eventcontent.header.messageto.split("-")[0]
        if message_target == "CGSRRouter":
            self.send_up(eventobj)


    def on_message_from_top(self, eventobj: Event):
        self.send_down(eventobj)



import sys
from ahc.Ahc import Event, EventTypes, GenericMessage, GenericMessageHeader, GenericMessagePayload,MessageDestinationIdentifiers, FramerObjects
from ahc.EttusUsrp.LiquidDspUtils import *
from ahc.EttusUsrp.FrameHandlerBase import FrameHandlerBase, framers, UsrpB210PhyEventTypes
from ctypes import *
import numpy as np
sys.path.append('/usr/local/lib')
import pickle
from threading import Lock
# framesync_callback = ctypes.CFUNCTYPE(ctypes.c_int32, ctypes.POINTER(ctypes.c_ubyte), ctypes.c_int32, ctypes.POINTER(ctypes.c_ubyte), ctypes.c_uint32, ctypes.c_int32, struct_c__SA_framesyncstats_s, ctypes.POINTER(None))

framers = FramerObjects()

mutex = Lock()
def ofdm_callback(header:POINTER(c_ubyte), header_valid:c_uint32, payload:POINTER(c_ubyte), payload_len:c_uint32, payload_valid:c_int32, stats:struct_c__SA_framesyncstats_s, userdata:POINTER(None)):
    mutex.acquire(1)
    try:
        framer = framers.get_framer_by_id(userdata)
        #print("ofdm_callback", framer.componentinstancenumber)
        #ofdmflexframegen_print(framer.fg)
        # userdata.debug_print()
        #print("Type", type(payload), "Payload Valid?: ", payload_valid, "Length=", payload_len, "payload=", bytes(payload))
        
        if payload_valid != 0:
            #ofdmflexframesync_print(framer.fs) 
            pload = string_at(payload, payload_len)
            #print("pload=", pload)
            phymsg = pickle.loads(pload)
            msg = GenericMessage(phymsg.header, phymsg.payload)
            framer.send_self(Event(framer, UsrpB210PhyEventTypes.RECV, msg))
            #print("Header=", msg.header.messagetype, " Payload=", msg.payload, " RSSI=", stats.rssi)
        #else:
            #pass
        #print("INVALID Type Node", framer.componentinstancenumber, "Payload Valid:[", payload_valid, "]Length=", payload_len, "payload=", bytes(payload))
    
    except Exception as e:
        print("Exception_ofdm_callback:", e)
        print("INVALID Type Node", framer.componentinstancenumber, "Payload Valid:[", payload_valid, "]Length=", payload_len, "payload=", bytes(payload))
    mutex.release()
    return 0

  
class UsrpB210OfdmFlexFramePhy(FrameHandlerBase):
    
    def on_init(self, eventobj: Event):
        #print("initialize LiquidDspOfdmFlexFrameHandler")
        pass
    
    def rx_callback(self, num_rx_samps, recv_buffer):
        try:
            #print("Self.fs", self.fs)
            ofdmflexframesync_execute(self.fs, recv_buffer.ctypes.data_as(POINTER(struct_c__SA_liquid_float_complex)) , num_rx_samps)
        except Exception as ex:
            print("Exception1", ex)

    
    def transmit(self, _header, _payload, _payload_len, _mod, _fec0, _fec1):
        #self.fgprops.mod_scheme = _mod
        #self.fgprops.fec0 = _fec0
        #self.fgprops.fec1 = _fec1
        #ofdmflexframegen_setprops(self.fg, byref(self.fgprops))
        ofdmflexframegen_assemble(self.fg, _header, _payload, _payload_len)
        # print("assembled")
        last_symbol = False
        while (last_symbol == 0):
            fgbuffer = np.zeros(self.fgbuffer_len, dtype=np.complex64)
            
            last_symbol = ofdmflexframegen_write(self.fg, fgbuffer.ctypes.data_as(POINTER(struct_c__SA_liquid_float_complex)), self.fgbuffer_len)
            #for i in range(self.fgbuffer_len):
            #    fgbuffer[i] = fgbuffer[i] * 2
            try:
                self.ahcuhd.transmit_samples(fgbuffer)
                # self.rx_callback(self.fgbuffer_len, npfgbuffer) #loopback for trial
            except Exception as e:
                print("Exception in transmit", e)
        self.ahcuhd.finalize_transmit_samples()
        #ofdmflexframesync_print(self.fs) 
            
        
    def configure(self):
        self.fgprops = ofdmflexframegenprops_s(LIQUID_CRC_32, LIQUID_FEC_NONE, LIQUID_FEC_HAMMING74, LIQUID_MODEM_QPSK)
        res = ofdmflexframegenprops_init_default(byref(self.fgprops))
        self.fgprops.check = LIQUID_CRC_32
        self.fgprops.fec0 = LIQUID_FEC_NONE
        self.fgprops.fec1 = LIQUID_FEC_HAMMING74
        self.fgprops.mod_scheme = LIQUID_MODEM_QPSK
        self.fgprops.M = 512
        self.fgprops.cp_len = 64
        self.fgprops.taper_len = 64
        self.fgbuffer_len = self.fgprops.M + self.fgprops.cp_len 

        self.fg = ofdmflexframegen_create(self.fgprops.M, self.fgprops.cp_len, self.fgprops.taper_len, None, byref(self.fgprops))

        res = ofdmflexframegen_print(self.fg)
        
        self.ofdm_callback_function = framesync_callback(ofdm_callback)
        
        try: 
            # WILL PASS ID of THIS OBJECT in userdata then will find the object in FramerObjects
            self.fs = ofdmflexframesync_create(self.fgprops.M, self.fgprops.cp_len, self.fgprops.taper_len, None, self.ofdm_callback_function, id(self))
            print("fs", self.fs, id(self))        
        except Exception as ex:
            print("Exception2", ex) 
        
        self.ahcuhd.start_rx(self.rx_callback, self)
        ofdmflexframegen_reset(self.fg)
        ofdmflexframesync_reset(self.fs)
        

# Callbacks have to be outside since the c library does not like "self"
# Because of this reason will use userdata to get access back to the framer object 

    def __init__(self, componentname, componentinstancenumber):
        super().__init__(componentname, componentinstancenumber)
        




import collections
import sys

from ahc.Ahc import ComponentModel, Event, ConnectorTypes, EventTypes, ComponentRegistry
import logging
logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

logger = logging.getLogger(__name__)

ProposalID = collections.namedtuple('ProposalID', ['number', 'uid'])

class PaxosMessage(object):

    from_uid = None  # Set by subclass constructor


class Prepare(PaxosMessage):


    def __init__(self, from_uid, proposal_id):
        self.from_uid = from_uid
        self.proposal_id = proposal_id


class NackPrepare(PaxosMessage):


    def __init__(self, from_uid, proposer_uid, proposal_id, promised_proposal_id):
        self.from_uid = from_uid
        self.proposal_id = proposal_id
        self.proposer_uid = proposer_uid
        self.promised_proposal_id = promised_proposal_id

class NackAccept(PaxosMessage):

    def __init__(self, from_uid, proposer_uid, proposal_id, promised_proposal_id):
        self.from_uid = from_uid
        self.proposal_id = proposal_id
        self.proposer_uid = proposer_uid
        self.promised_proposal_id = promised_proposal_id

class Promise(PaxosMessage):


    def __init__(self, from_uid, proposer_uid, proposal_id, last_accepted_id, last_accepted_value):
        self.from_uid = from_uid
        self.proposer_uid = proposer_uid
        self.proposal_id = proposal_id
        self.last_accepted_id = last_accepted_id
        self.last_accepted_value = last_accepted_value


class Accept(PaxosMessage):

    def __init__(self, from_uid, proposal_id, proposal_value):
        self.from_uid = from_uid
        self.proposal_id = proposal_id
        self.proposal_value = proposal_value


class Accepted(PaxosMessage):

    def __init__(self, from_uid, proposal_id, proposal_value):
        self.from_uid = from_uid
        self.proposal_id = proposal_id
        self.proposal_value = proposal_value


class Resolution(PaxosMessage):

    def __init__(self, from_uid, value):
        self.from_uid = from_uid
        self.value = value


class InvalidMessageError(Exception):
    pass



class Proposer(object):

    leader = False
    proposed_value = None
    proposal_id = ProposalID(0, ' ')
    highest_accepted_id = ProposalID(0, ' ')
    promises_received = None
    nacks_received = None
    current_prepare_msg = None
    current_accept_msg = None

    def __init__(self, network_uid, quorum_size):
        self.network_uid = network_uid
        self.quorum_size = quorum_size
        self.proposal_id = ProposalID(0, network_uid)
        self.highest_proposal_id = ProposalID(0, network_uid)

    def propose_value(self, value):

        if self.proposed_value is None:
            self.proposed_value = value

            if self.leader:
                self.current_accept_msg = Accept(self.network_uid, self.proposal_id, value)
                return self.current_accept_msg
            else:
                return None

        else:
            return None

    def prepare(self):

        self.leader = False
        self.promises_received = set()
        self.nacks_received = set()
        self.proposal_id = ProposalID(self.highest_proposal_id.number + 1, self.network_uid)
        self.highest_proposal_id = self.proposal_id
        self.current_prepare_msg = Prepare(self.network_uid, self.proposal_id)
        return self.current_prepare_msg

    def observe_proposal(self, proposal_id):

        if proposal_id > self.highest_proposal_id:
            self.highest_proposal_id = proposal_id

    def receive_nack(self, msg):
        logger.info("For component %s NACK received from %s", self.network_uid.componentinstancenumber,
                    msg.from_uid.componentinstancenumber)
        self.observe_proposal(msg.promised_proposal_id)

        if msg.proposal_id == self.proposal_id and self.nacks_received is not None:
            self.nacks_received.add(msg.from_uid)

            if len(self.nacks_received) == self.quorum_size-1:
                return self.prepare()  # Lost leadership or failed to acquire it

    def receive_promise(self, msg):
        logger.info("For component %s Promise received from %s", self.network_uid.componentinstancenumber,
                    msg.from_uid.componentinstancenumber)
        self.observe_proposal(msg.proposal_id)

        if not self.leader and msg.proposal_id == self.proposal_id and msg.from_uid not in self.promises_received:

            self.promises_received.add(msg.from_uid)

            if msg.last_accepted_id is not None and msg.last_accepted_id > self.highest_accepted_id:
                self.highest_accepted_id = msg.last_accepted_id
                if msg.last_accepted_value is not None:
                    self.proposed_value = msg.last_accepted_value
            # apart from self
            if len(self.promises_received) == self.quorum_size-1:
                self.leader = True

                if self.proposed_value is not None:
                    self.current_accept_msg = Accept(self.network_uid, self.proposal_id, self.proposed_value)
                    return self.current_accept_msg


class Acceptor(object):

    def __init__(self, network_uid, promised_id=None, accepted_id=None, accepted_value=None):

        self.network_uid = network_uid
        self.promised_id = promised_id
        self.accepted_id = accepted_id
        self.accepted_value = accepted_value

    def receive_prepare(self, msg):
        logger.info("For component %s Prepare received from %s", self.network_uid.componentinstancenumber,
                    msg.from_uid.componentinstancenumber)
        if msg is None or msg.proposal_id is None:
            return NackPrepare(self.network_uid, msg.from_uid, msg.proposal_id, self.promised_id)

        if self.promised_id is None or msg.proposal_id >= self.promised_id:
            self.promised_id = msg.proposal_id
            return Promise(self.network_uid, msg.from_uid, self.promised_id, self.accepted_id, self.accepted_value)
        else:
            return NackPrepare(self.network_uid, msg.from_uid, msg.proposal_id, self.promised_id)

    def receive_accept(self, msg):

        logger.info("For component %s Accept received from %s", self.network_uid.componentinstancenumber,
                    msg.from_uid.componentinstancenumber)
        if msg is None or msg.proposal_id is None:
            return NackAccept(self.network_uid, msg.from_uid, msg.proposal_id, self.promised_id)
        if self.promised_id is None or msg.proposal_id >= self.promised_id:
            self.promised_id = msg.proposal_id
            self.accepted_id = msg.proposal_id
            self.accepted_value = msg.proposal_value
            return Accepted(self.network_uid, msg.proposal_id, msg.proposal_value)
        else:
            return NackAccept(self.network_uid, msg.from_uid, msg.proposal_id, self.promised_id)


class Learner(object):


    class ProposalStatus(object):
        __slots__ = ['accept_count', 'retain_count', 'acceptors', 'value']

        def __init__(self, value):
            self.accept_count = 0
            self.retain_count = 0
            self.acceptors = set()
            self.value = value

    def __init__(self, network_uid, quorum_size):
        self.network_uid = network_uid
        self.quorum_size = quorum_size
        self.proposals = dict()  # maps proposal_id => ProposalStatus
        self.acceptors = dict()  # maps from_uid => last_accepted_proposal_id
        self.final_value = None
        self.final_acceptors = None  # Will be a set of acceptor UIDs once the final value is chosen
        self.final_proposal_id = ProposalID(0, ' ')

    def receive_accepted(self, msg):
        logger.info("For component %s AcceptED received from %s", self.network_uid.componentinstancenumber,
                    msg.from_uid.componentinstancenumber)
        if self.final_value is not None:
            if msg.proposal_id >= self.final_proposal_id and msg.proposal_value == self.final_value:
                self.final_acceptors.add(msg.from_uid)
            return Resolution(self.network_uid, self.final_value)

        last_pn = self.acceptors.get(msg.from_uid)

        if last_pn is not None and msg.proposal_id <= last_pn:
            return  # Old message

        self.acceptors[msg.from_uid] = msg.proposal_id

        if last_pn is not None:
            ps = self.proposals[last_pn]
            ps.retain_count -= 1
            ps.acceptors.remove(msg.from_uid)
            if ps.retain_count == 0:
                del self.proposals[last_pn]

        if not msg.proposal_id in self.proposals:
            self.proposals[msg.proposal_id] = Learner.ProposalStatus(msg.proposal_value)

        ps = self.proposals[msg.proposal_id]

        assert msg.proposal_value == ps.value, 'Value mismatch for single proposal!'

        ps.accept_count += 1
        ps.retain_count += 1
        ps.acceptors.add(msg.from_uid)

        if ps.accept_count >= self.quorum_size/2:
            self.final_proposal_id = msg.proposal_id
            self.final_value = msg.proposal_value
            self.final_acceptors = ps.acceptors
            self.proposals = None
            self.acceptors = None

            return Resolution(self.network_uid, self.final_value)


class PaxosInstance(Proposer, Acceptor, Learner):

    def __init__(self, network_uid, quorum_size, promised_id=ProposalID(0, ' '), accepted_id=ProposalID(0, ' '),
                 accepted_value=None):
        Proposer.__init__(self, network_uid, quorum_size)
        Acceptor.__init__(self, network_uid, promised_id, accepted_id, accepted_value)
        Learner.__init__(self, network_uid, quorum_size)

    def receive_prepare(self, msg):
        self.observe_proposal(msg.proposal_id)
        return super(PaxosInstance, self).receive_prepare(msg)

    def receive_accept(self, msg):
        self.observe_proposal(msg.proposal_id)
        return super(PaxosInstance, self).receive_accept(msg)


class PaxosConsensusComponentModel(ComponentModel):

    def __init__(self, componentname, componentinstancenumber):
        super().__init__(componentname, componentinstancenumber)
        self.acceptor: Acceptor = None
        self.proposer: Proposer = None
        self.learner: Learner = None
        self.client = None
        self.cluster_size = 1

    def on_message_from_bottom(self, eventobj: Event):
        self.data_received_peer(eventobj.eventsource, eventobj.eventcontent)

    def on_message_from_top(self, eventobj: Event):
        print("client asking for a trouble...")
        self.data_received_client(eventobj.eventsource, eventobj.eventcontent)

    def data_received_peer(self, sender, message):
        if isinstance(message, Prepare):
            result_prepare = self.acceptor.receive_prepare(message)
            self.send_to_component(sender, result_prepare)
        elif isinstance(message, Promise):
            result_promise = self.proposer.receive_promise(message)
            if result_promise is not None:
                self.broadcast_peers(result_promise)

        elif isinstance(message, NackAccept) or isinstance(message, NackPrepare):
            result = self.proposer.receive_nack(message)
            if result is not None:
                self.broadcast_peers(result)
        elif isinstance(message, Accept):
            self.proposer.observe_proposal(message.proposal_id)
            result = self.acceptor.receive_accept(message)
            if result is not None:
                self.broadcast_peers(result)
        elif isinstance(message, Accepted):
            result = self.learner.receive_accepted(message)
            if result is not None:
                self.broadcast_peers(result)
        elif isinstance(message, Resolution):
            if self.client is not None:
                self.send(self.client, message)

    def data_received_client(self, client, message):
        self.client = client
        prep_message = self.proposer.prepare()
        self.broadcast_peers(prep_message)
        proposal = self.proposer.propose_value(message)
        if proposal is not None:
            self.broadcast_peers(proposal)

    def send(self, client, message:Resolution):
        client.send(message)

    def send_to_component(self, recipient, message):
        if recipient != self:
            for conn in self.connectors[ConnectorTypes.DOWN]:
                if conn.componentinstancenumber.find(recipient.componentinstancenumber) != -1:
                    conn.trigger_event(Event(self, EventTypes.MFRT, message))

    def broadcast_peers(self, message):
        self.send_down(Event(self, EventTypes.MFRT, message))

    def on_init(self, eventobj: Event):
        self.cluster_size = len(self.registry.get_non_channel_components())
        self.acceptor = Acceptor(self, ProposalID(0, self.componentinstancenumber), None, None)
        self.proposer = Proposer(self, self.cluster_size)
        self.learner = Learner(self, self.cluster_size)
